# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from rich.console import Console
console = Console()



# Copyright (C) 2025 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)
# Copyright 2025 BSD 3-Clause License (see https://opensource.org/license/bsd-3-clause)

# type: ignore
####  IMDB imports ###
import json
import random
import urllib3
import requests
from urllib3.exceptions import InsecureRequestWarning
from .sdxclasses import GenerateUserAgent

urllib3.disable_warnings(InsecureRequestWarning)

### html_requests imports ###
import sys
import asyncio
from urllib.parse import urlparse, urlunparse, urljoin
from typing import Set, Union, List, MutableMapping
from pyquery import PyQuery
from lxml.html.clean import Cleaner
import lxml
from lxml import etree
from lxml.html import HtmlElement
from lxml.html import tostring as lxml_html_tostring
from lxml.html.soupparser import fromstring as soup_parse
from parse import search as parse_search
from parse import findall, Result
from w3lib.encoding import html_to_unicode

__all__ = ["IMDB"]

### IMDB search classes ###
class ImdbParser:
    """
      - A class to manipulate incoming json string data of a movie/TV from IMDB.
      - Changes are required as sometimes the json contains invalid chars in description/reviewBody/trailer schema
    """
    def __init__(self, json_string):
        self.json_string = json_string

    @property
    def remove_trailer(self):
        """
         @description:- Helps to remove 'trailer' schema from IMDB data json string.
         @returns:- New updated JSON string.
        """
        try:
            self.json_string = ''.join(self.json_string.splitlines())
            trailer_i = self.json_string.index('"trailer"')
            actor_i = self.json_string.index('"actor"')
            to_remove = self.json_string[trailer_i:actor_i:1]
            self.json_string = self.json_string.replace(to_remove, "")
        except ValueError:
            self.json_string = self.json_string
        return self.json_string

    @property
    def remove_description(self):
        """
         @description:- Helps to remove 'description' schema from IMDB file json string.
         @returns:- New updated JSON string.
        """
        try:
            review_i = self.json_string.index('"review"')
            des_i = self.json_string.index('"description"', 0, review_i)
            to_remove = self.json_string[des_i:review_i:1]
            self.json_string = self.json_string.replace(to_remove, "")
        except ValueError:
            self.json_string = self.json_string
        return self.json_string

    @property
    def remove_review_body(self):
        """
         @description:- Helps to remove 'reviewBody' schema from IMDB file json string.
         @returns:- New updated JSON string.
        """
        try:
            reviewrating_i = self.json_string.index('"reviewRating"')
            reviewbody_i = self.json_string.index('"reviewBody"', 0, reviewrating_i)
            to_remove = self.json_string[reviewbody_i:reviewrating_i:1]
            self.json_string = self.json_string.replace(to_remove, "")
        except ValueError:
            self.json_string = self.json_string
        return self.json_string

class IMDB:
    """
        A class to represent IMDB API.

        --------------

        Main Methods of the IMDB API
        --------------
            #1. search(name, year=None, tv=False)
                -- to search a query on IMDB

            #2. get_by_name(name, year=None, tv=False)
                -- to get a Movie/TV-Series info by it's name (pass year also to increase accuracy)
    """
    def __init__(self):
        self.session = HTMLSession()
        ua = GenerateUserAgent.random_browser()
        self.headers = {
           "Accept": "application/json, text/plain, */*",
           "Accept-Language": "es-ES,es,q=0.6",
           "User-Agent": ua,
           "Referer": "https://www.imdb.com/"
           }
        self.baseURL = "https://www.imdb.com"
        self.search_results = {'result_count': 0, 'results': []}
        self.NA = json.dumps({"status": 404, "message": "No Result Found!", 'result_count': 0, 'results': []})

    # ..................................method to search on IMDB...........................................
    def search(self, name, year=None, tv=False):
        """
         @description:- Helps to search a query on IMDB.
         @parameter-1:- <str:name>, query value to search.
         @parameter-2:- <int:year> OPTIONAL, release year of query/movie/tv/file to search.
         @parameter-3:- <bool:tv> OPTIONAL, to filter/limit/bound search results only for 'TV Series'.
         @returns:- A JSON string:
                    - {'result_count': <int:total_search_results>, 'results': <list:list_of_files/movie_info_dict>}
        """
        assert isinstance(name, str)
        self.search_results = {'result_count': 0, 'results': []}

        name = name.replace(" ", "+")

        if year is None:
            url = f"https://www.imdb.com/find?q={name}"
        else:
            assert isinstance(year, int)
            url = f"https://www.imdb.com/find?q={name}+{year}"

        try:
            response = self.session.get(url)
        except requests.exceptions.ConnectionError as e:
            response = self.session.get(url, verify=False)

        results = response.html.xpath("//section[@data-testid='find-results-section-title']/div/ul/li")
        if tv is True:
            results = [result for result in results if "TV" in result.text]
        else:
            results = [result for result in results if "TV" not in result.text]
        
        output = []
        for result in results:
            name = result.text.replace('\n', ' ')
            year_date = "N/A"
            show = ""
            url = result.find('a')[0].attrs['href']
            if not (any(s in name for s in ['Podcast', 'Music Video', 'Video', 'Episode', 'Short'])):
                try:
                    for i in range(len(result.find('span'))):
                        span = result.find('span')[i]
                        if 'TV' in span.text:
                            show = span.text
                        else:
                            text = span.text.strip().split('-')[0][:4]
                            if text.isnumeric():
                                year_date = text
                          
                    show = "Movie" if show == "" else show
            
                    file_id = url.split('/')[2]
                    name = result.find('a')[0].text
                    output.append({
                        "type": show,
                        "id": file_id,
                        "year": year_date,
                        "name": name,
                        "url": f"https://www.imdb.com{url}"
                       })
                except IndexError:
                    pass
                self.search_results = {'result_count': len(output), 'results': output}
        return json.dumps(self.search_results, indent=2)

    # ..............................methods to get a movie/web-series/tv info..............................
    def get(self, url):
        """
         @description:- helps to get a file's complete info (used by get_by_name() & get_by_id() )
         @parameter:- <str:url>, url of the file/movie/tv-series.
         @returns:- File/movie/TV info as JSON string.
        """
        try:
            response = self.session.get(url)
            result = response.html.xpath("//script[@type='application/ld+json']")[0].text
            result = ''.join(result.splitlines())  # removing newlines
            result = f"""{result}"""

        except IndexError:
            return self.NA
        try:
            # converting json string into dict
            result = json.loads(result)
        except json.decoder.JSONDecodeError as e:
            # sometimes json is invalid as 'description' contains inverted commas or other html escape chars
            try:
                to_parse = ImdbParser(result)
                # removing trailer & description schema from json string
                parsed = to_parse.remove_trailer
                # parsed = to_parse.remove_description

                result = json.loads(parsed)
            except json.decoder.JSONDecodeError as e:
                try:
                    # removing reviewBody from json string
                    parsed = to_parse.remove_review_body 
                    result = json.loads(parsed)
                except json.decoder.JSONDecodeError as e:
                    # invalid char(s) is/are not in description/trailer/reviewBody schema
                    return self.NA
        
        output = {
            "type": result.get('@type'),
            "id":result.get('url').split(self.baseURL + "/title")[-1].strip("/"),
            "name": result.get('name'),
            "year": str(result.get("datePublished"))[:-6],
            "url": result.get('url'),
            "description": result.get('description')
        }
        return json.dumps(output, indent=2)

    def get_by_name(self, name, year=None, tv=False):
        """
         @description:- Helps to search a file/movie/tv by name.
         @parameter-1:- <str:name>, query/name to search.
         @parameter-2:- <int:year> OPTIONAL, release year of query/movie/tv/file to search.
         @parameter-3:- <bool:tv> OPTIONAL, to filter/limit/bound search result only for 'TV Series'.
         @returns:- File/movie/TV info as JSON string.
        """
        results = json.loads(self.search(name, year=year, tv=tv))
        all_results = [i for i in self.search_results['results'] if 'title' in i['url']]

        # filtering TV and movies
        if tv is True:  # for tv/Web-Series only
            tv_only = [result for result in all_results if "TV" in result['type']]
            if year is not None:
                tv_only = [result for result in tv_only if str(year) in result['name']]
            # double checking by file name
            if bool(tv_only):
                tv_only_checked = [result for result in tv_only if result['name'].lower().startswith(name.split(" ")[0].lower())]
                tv_only = tv_only_checked if bool(tv_only_checked) else tv_only
            results['results'] = tv_only if bool(tv_only) else all_results

        else:  # for movies only
            movie_only = [result for result in all_results if "TV" not in result['name']]
            if year is not None:
                movie_only = [result for result in movie_only if str(year) in result['name']]
            # double checking by file name
            if bool(movie_only):
                movie_only_checked = [result for result in movie_only if result['name'].lower().startswith(name.split(" ")[0].lower())]
                movie_only = movie_only_checked if bool(movie_only_checked) else movie_only
            results['results'] = movie_only if bool(movie_only) else all_results

        if len(results['results']) > 0:
            return self.get(results['results'][0].get('url'))
        else:
            return self.NA

    def get_by_id(self, file_id):
        """
         @description:- Helps to search a file/movie/tv by its imdb ID.
         @parameter-1:- <str:file_id>, imdb ID of the file/movie/tv.
         @returns:- File/movie/TV info as JSON string.
        """
        assert isinstance(file_id, str)
        url = f"{self.baseURL}/title/{file_id}"
        return self.get(url)
    
### html_requests classes ###
DEFAULT_ENCODING = 'utf-8'
DEFAULT_URL = 'https://example.org/'
DEFAULT_USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/603.3.8 (KHTML, like Gecko) Version/10.1.2 Safari/603.3.8'
DEFAULT_NEXT_SYMBOL = ['next', 'more', 'older']

cleaner = Cleaner(javascript=True, style=True)
# cleaner.javascript = True
# cleaner.style = True

useragent = None

# Typing.
_Find = Union[List['Element'], 'Element']
_XPath = Union[List[str], List['Element'], str, 'Element']
_Result = Union[List['Result'], 'Result']
_HTML = Union[str, bytes]
_BaseHTML = str
_UserAgent = str
_DefaultEncoding = str
_URL = str
_RawHTML = bytes
_Encoding = str
_Text = str
_Containing = Union[str, List[str]]
_Links = Set[str]
_Attrs = MutableMapping
_Next = Union['HTML', List[str]]
_NextSymbol = List[str]

# Sanity checking.
try:
    assert sys.version_info.major == 3
    assert sys.version_info.minor > 5
except AssertionError:
    raise RuntimeError('Requests-HTML requires Python 3.6+!')


class MaxRetries(Exception):

    def __init__(self, message):
        self.message = message


class BaseParser:
    """A basic HTML/Element Parser, for Humans.

    :param element: The element from which to base the parsing upon.
    :param default_encoding: Which encoding to default to.
    :param html: HTML from which to base the parsing upon (optional).
    :param url: The URL from which the HTML originated, used for ``absolute_links``.

    """

    def __init__(self, *, element, default_encoding: _DefaultEncoding = "", html: _HTML = "", url: _URL) -> None:
        self.element = element
        self.url = url
        self.skip_anchors = True
        self.default_encoding = default_encoding
        self._encoding = None
        self._html = html.encode(DEFAULT_ENCODING) if isinstance(html, str) else html
        self._lxml = None
        self._pq = None

    @property
    def raw_html(self) -> _RawHTML:
        """Bytes representation of the HTML content.
        (`learn more <http://www.diveintopython3.net/strings.html>`_).
        """
        if self._html:
            return self._html
        else:
            return etree.tostring(self.element, encoding='unicode').strip().encode(self.encoding) 

    @property
    def html(self) -> _BaseHTML:
        """Unicode representation of the HTML content
        (`learn more <http://www.diveintopython3.net/strings.html>`_).
        """
        if self._html:
            return self.raw_html.decode(self.encoding, errors='replace')
        else:
            return etree.tostring(self.element, encoding='unicode').strip() 

    @html.setter
    def html(self, html: str) -> None:
        self._html = html.encode(self.encoding)

    @raw_html.setter
    def raw_html(self, html: bytes) -> None:
        """Property setter for self.html."""
        self._html = html

    @property
    def encoding(self) -> _Encoding:
        """The encoding string to be used, extracted from the HTML and
        :class:`HTMLResponse <HTMLResponse>` headers.
        """
        if self._encoding:
            return self._encoding

        # Scan meta tags for charset.
        if self._html:
            self._encoding = html_to_unicode(self.default_encoding, self._html)[0]
            # Fall back to requests' detected encoding if decode fails.
            try:
                self.raw_html.decode(self.encoding, errors='replace')
            except UnicodeDecodeError:
                self._encoding = self.default_encoding


        return self._encoding if self._encoding else self.default_encoding

    @encoding.setter
    def encoding(self, enc: str) -> None:
        """Property setter for self.encoding."""
        self._encoding = enc

    @property
    def pq(self) -> PyQuery:
        """`PyQuery <https://pythonhosted.org/pyquery/>`_ representation
        of the :class:`Element <Element>` or :class:`HTML <HTML>`.
        """
        if self._pq is None:
            self._pq = PyQuery(self.lxml)

        return self._pq

    @property
    def lxml(self) -> HtmlElement:
        """`lxml <http://lxml.de>`_ representation of the
        :class:`Element <Element>` or :class:`HTML <HTML>`.
        """
        if self._lxml is None:
            try:
                self._lxml = soup_parse(self.html, features='html.parser')
            except ValueError:
                self._lxml = lxml.html.fromstring(self.raw_html) 

        return self._lxml

    @property
    def text(self) -> _Text:
        """The text content of the
        :class:`Element <Element>` or :class:`HTML <HTML>`.
        """
        return str(self.pq.text())

    @property
    def full_text(self) -> _Text:
        """The full text content (including links) of the
        :class:`Element <Element>` or :class:`HTML <HTML>`.
        """
        return self.lxml.text_content()

    def find(self, selector: str = "*", *, containing: _Containing = "", clean: bool = False, first: bool = False, _encoding: str = "") -> _Find:
        """Given a CSS Selector, returns a list of
        :class:`Element <Element>` objects or a single one.

        :param selector: CSS Selector to use.
        :param clean: Whether or not to sanitize the found HTML of ``<script>`` and ``<style>`` tags.
        :param containing: If specified, only return elements that contain the provided text.
        :param first: Whether or not to return just the first result.
        :param _encoding: The encoding format.

        Example CSS Selectors:

        - ``a``
        - ``a.someClass``
        - ``a#someID``
        - ``a[target=_blank]``

        See W3School's `CSS Selectors Reference
        <https://www.w3schools.com/cssref/css_selectors.asp>`_
        for more details.

        If ``first`` is ``True``, only returns the first
        :class:`Element <Element>` found.
        """

        # Convert a single containing into a list.
        if isinstance(containing, str):
            containing = [containing]

        encoding = _encoding or self.encoding
        elements = [
            Element(element=found, url=self.url, default_encoding=encoding)
            for found in self.pq(selector)
        ]

        if containing:
            elements_copy = elements.copy()
            elements = []

            for element in elements_copy:
                if any([c.lower() in element.full_text.lower() for c in containing]):
                    elements.append(element)

            elements.reverse()

        # Sanitize the found HTML.
        if clean:
            elements_copy = elements.copy()
            elements = []

            for element in elements_copy:
                element.raw_html = lxml_html_tostring(cleaner.clean_html(element.lxml)) 
                elements.append(element)

        return _get_first_or_list(elements, first) 

    def xpath(self, selector: str, *, clean: bool = False, first: bool = False, _encoding: str = "") -> _XPath:
        """Given an XPath selector, returns a list of
        :class:`Element <Element>` objects or a single one.

        :param selector: XPath Selector to use.
        :param clean: Whether or not to sanitize the found HTML of ``<script>`` and ``<style>`` tags.
        :param first: Whether or not to return just the first result.
        :param _encoding: The encoding format.

        If a sub-selector is specified (e.g. ``//a/@href``), a simple
        list of results is returned.

        See W3School's `XPath Examples
        <https://www.w3schools.com/xml/xpath_examples.asp>`_
        for more details.

        If ``first`` is ``True``, only returns the first
        :class:`Element <Element>` found.
        """
        selected = self.lxml.xpath(selector)

        elements = [
            Element(element=selection, url=self.url, default_encoding=_encoding or self.encoding)
            if not isinstance(selection, etree._ElementUnicodeResult) else str(selection)
            for selection in selected
        ]

        # Sanitize the found HTML.
        if clean:
            elements_copy = elements.copy()
            elements = []

            for element in elements_copy:
                element.raw_html = lxml_html_tostring(cleaner.clean_html(element.lxml)) 
                elements.append(element)

        return _get_first_or_list(elements, first) 

    def search(self, template: str) -> Result:
        """Search the :class:`Element <Element>` for the given Parse template.

        :param template: The Parse template to use.
        """

        return parse_search(template, self.html) 

    def search_all(self, template: str) -> _Result:
        """Search the :class:`Element <Element>` (multiple times) for the given parse
        template.

        :param template: The Parse template to use.
        """
        return [r for r in findall(template, self.html)] 

    @property
    def links(self) -> _Links:
        """All found links on page, in as–is form."""

        def gen():
            for link in self.find('a'): 

                try:
                    href = link.attrs['href'].strip()
                    if href and not (href.startswith('#') and self.skip_anchors) and not href.startswith(('javascript:', 'mailto:')):
                        yield href
                except KeyError:
                    pass

        return set(gen())

    def _make_absolute(self, link):
        """Makes a given link absolute."""

        # Parse the link with stdlib.
        parsed = urlparse(link)._asdict()

        # If link is relative, then join it with base_url.
        if not parsed['netloc']:
            return urljoin(self.base_url, link)

        # Link is absolute; if it lacks a scheme, add one from base_url.
        if not parsed['scheme']:
            parsed['scheme'] = urlparse(self.base_url).scheme

            # Reconstruct the URL to incorporate the new scheme.
            parsed = (v for v in parsed.values())
            return urlunparse(parsed)

        # Link is absolute and complete with scheme; nothing to be done here.
        return link

    @property
    def absolute_links(self) -> _Links:
        """All found links on page, in absolute form
        (`learn more <https://www.navegabem.com/absolute-or-relative-links.html>`_).
        """

        def gen():
            for link in self.links:
                yield self._make_absolute(link)

        return set(gen()) 

    @property
    def base_url(self) -> _URL:
        """The base URL for the page. Supports the ``<base>`` tag
        (`learn more <https://www.w3schools.com/tags/tag_base.asp>`_)."""

        # Support for <base> tag.
        base = self.find('base', first=True)
        if base:
            result = base.attrs.get('href', '').strip() 
            if result:
                return result

        # Parse the url to separate out the path
        parsed = urlparse(self.url)._asdict()

        # Remove any part of the path after the last '/'
        parsed['path'] = '/'.join(parsed['path'].split('/')[:-1]) + '/'

        # Reconstruct the url with the modified path
        parsed = (v for v in parsed.values())
        url = urlunparse(parsed)

        return str(url)

class Element(BaseParser):
    """An element of HTML.

    :param element: The element from which to base the parsing upon.
    :param url: The URL from which the HTML originated, used for ``absolute_links``.
    :param default_encoding: Which encoding to default to.
    """

    __slots__ = [
        'element', 'url', 'skip_anchors', 'default_encoding', '_encoding',
        '_html', '_lxml', '_pq', '_attrs', 'session'
    ]

    def __init__(self, *, element, url: _URL, default_encoding: _DefaultEncoding = "") -> None:
        super(Element, self).__init__(element=element, url=url, default_encoding=default_encoding)
        self.element = element
        self.tag = element.tag
        self.lineno = element.sourceline
        self._attrs = None

    def __repr__(self) -> str:
        attrs = ['{}={}'.format(attr, repr(self.attrs[attr])) for attr in self.attrs]
        return "<Element {} {}>".format(repr(self.element.tag), ' '.join(attrs))

    @property
    def attrs(self) -> _Attrs:
        """Returns a dictionary of the attributes of the :class:`Element <Element>`
        (`learn more <https://www.w3schools.com/tags/ref_attributes.asp>`_).
        """
        if self._attrs is None:
            self._attrs = {k: v for k, v in self.element.items()}

            # Split class and rel up, as there are ussually many of them:
            for attr in ['class', 'rel']:
                if attr in self._attrs:
                    self._attrs[attr] = tuple(self._attrs[attr].split())

        return self._attrs

class HTML(BaseParser):
    """An HTML document, ready for parsing.

    :param url: The URL from which the HTML originated, used for ``absolute_links``.
    :param html: HTML from which to base the parsing upon (optional).
    :param default_encoding: Which encoding to default to.
    """

    def __init__(self, *, session: Union['HTMLSession', 'None'], url: str = DEFAULT_URL, html: _HTML, default_encoding: str = DEFAULT_ENCODING) -> None:

        # Convert incoming unicode HTML into bytes.
        if isinstance(html, str):
            html = html.encode(DEFAULT_ENCODING)

        super(HTML, self).__init__(
            # Convert unicode HTML to bytes.
            element=PyQuery(html)('html') or PyQuery(f'<html>{html}</html>')('html'),
            html=html,
            url=url,
            default_encoding=default_encoding
        )
        self.session = session or HTMLSession()
        self.page = None
        self.next_symbol = DEFAULT_NEXT_SYMBOL

    def __repr__(self) -> str:
        return f"<HTML url={self.url!r}>"

    def next(self, fetch: bool = False, next_symbol: _NextSymbol = DEFAULT_NEXT_SYMBOL) -> _Next:
        """Attempts to find the next page, if there is one. If ``fetch``
        is ``True`` (default), returns :class:`HTML <HTML>` object of
        next page. If ``fetch`` is ``False``, simply returns the next URL.

        """

        def get_next():
            candidates = self.find('a', containing=next_symbol)

            for candidate in candidates: 
                if candidate.attrs.get('href'):
                    # Support 'next' rel (e.g. reddit).
                    if 'next' in candidate.attrs.get('rel', []):
                        return candidate.attrs['href']

                    # Support 'next' in classnames.
                    for _class in candidate.attrs.get('class', []):
                        if 'next' in _class:
                            return candidate.attrs['href']

                    if 'page' in candidate.attrs['href']:
                        return candidate.attrs['href']

            try:
                # Resort to the last candidate.
                return candidates[-1].attrs['href'] 
            except IndexError:
                return None

        __next = get_next()
        if __next:
            url = self._make_absolute(__next)
        else:
            return None 

        if fetch:
            return self.session.get(url) 
        else:
            return url 

    def __iter__(self):

        next = self

        while True:
            yield next
            try:
                next = next.next(fetch=True, next_symbol=self.next_symbol).html 
            except AttributeError:
                break

    def __next__(self):
        return self.next(fetch=True, next_symbol=self.next_symbol).html 

    def __aiter__(self):
        return self

    def add_next_symbol(self, next_symbol):
        self.next_symbol.append(next_symbol)

class HTMLResponse(requests.Response):
    """An HTML-enabled :class:`requests.Response <requests.Response>` object.
    Effectively the same, but with an intelligent ``.html`` property added.
    """

    def __init__(self, session: Union['HTMLSession','HTMLSession']) -> None:
        super(HTMLResponse, self).__init__()
        self._html = None   # type: HTML
        self.session = session

    @property
    def html(self) -> HTML:
        if not self._html:
            self._html = HTML(session=self.session, url=self.url, html=self.content, default_encoding=self.encoding) 

        return self._html

    @classmethod
    def _from_response(cls, response, session: Union['HTMLSession', 'HTMLSession']):
        html_r = cls(session=session)
        html_r.__dict__.update(response.__dict__)
        return html_r

def user_agent(style=None) -> _UserAgent:
    """Returns an apparently legit user-agent, if not requested one of a specific
    style. Defaults to a Chrome-style User-Agent.
    """
    global useragent
    ua = GenerateUserAgent
    browsers = {
        'chrome'  : random.choice(ua.chrome()),
        'firefox' : random.choice(ua.firefox()),
        'opera'   : random.choice(ua.opera())
    }

    if (not useragent) and style:
        useragent = browsers.get(style)
    else:
        useragent = ua.random_browser()

    return useragent if useragent else DEFAULT_USER_AGENT

def _get_first_or_list(l, first=False):
    if first:
        try:
            return l[0]
        except IndexError:
            return None
    else:
        return l

class BaseSession(requests.Session):
    """ A consumable session, for cookie persistence and connection pooling,
    amongst other things.
    """

    def __init__(self, mock_browser : bool = True, verify : bool = True,
                 browser_args : list = ['--no-sandbox']):
        super().__init__()

        # Mock a web browser's user agent.
        if mock_browser:
            self.headers['User-Agent'] = user_agent()
        
        self.hooks['response'].append(self.response_hook)
        self.verify = verify

        self.__browser_args = browser_args

    def response_hook(self, response, **kwargs) -> HTMLResponse:
        """ Change response enconding and replace it by a HTMLResponse. """
        if not response.encoding:
            response.encoding = DEFAULT_ENCODING
        return HTMLResponse._from_response(response, self) 

class HTMLSession(BaseSession):

    def __init__(self, **kwargs):
        super(HTMLSession, self).__init__(**kwargs)

    @property
    def browser(self):
        if not hasattr(self, "_browser"):
            self.loop = asyncio.get_event_loop()
            if self.loop.is_running():
                raise RuntimeError("Cannot use HTMLSession within an existing event loop. Use AsyncHTMLSession instead.")
            self._browser = self.loop.run_until_complete(super().browser) 
        return self._browser

    def close(self):
        """ If a browser was created close it first. """
        if hasattr(self, "_browser"):
            self.loop.run_until_complete(self._browser.close())
        super().close()


#!/bin/env python
# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
from typing import Dict, Any
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl
from sdx_dl.sdxparser import logger, args as parser_args
from sdx_dl.sdxlib import get_subtitle_id, get_subtitle
from sdx_dl.sdxutils import sub_extensions, Metadata as Metadata, extract_meta_data
from sdx_dl.sdxclasses import FindFiles, NoResultsError, VideoMetadataExtractor
from guessit import guessit # type: ignore
from contextlib import contextmanager

_extensions = [
    'avi', 'mkv', 'mp4',
    'mpg', 'm4v', 'ogv',
    'vob', '3gp',
    'part', 'temp', 'tmp'
]

@contextmanager
def subtitle_renamer(filepath:str, inf_sub:Dict[str, Any]):
    """Dectect new subtitles files in a directory and rename with
       filepath basename."""

    def extract_name(filepath:str):
        """.Extract Filename."""
        filename, fileext = os.path.splitext(filepath)
        if fileext in ('.part', '.temp', '.tmp'):
            filename, fileext = os.path.splitext(filename)
        return filename
   
    dirpath = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    before = set(os.listdir(dirpath))
    yield
    after = set(os.listdir(dirpath))

    # Fixed error for rename various subtitles with same filename
    info = {}
    for new_file in after - before:
        new_ext = os.path.splitext(new_file)[1]
        if new_ext not in sub_extensions:
            # only apply to subtitles
            continue
        filename = extract_name(filepath)
        new_file_dirpath = os.path.join(os.path.dirname(filename), new_file)

        try:
           if os.path.exists(filename + new_ext):
               continue
           else:
                if inf_sub['type'] == "episode" and inf_sub['season']:
                    info = guessit(new_file) # type: ignore
                    number = f"s{info['season']:02}e{info['episode']:02}" if "season" in info and "episode" in info else None
                    if number == inf_sub['number']:
                        os.rename(new_file_dirpath, filename + new_ext)
                    else:
                        continue
                else:
                    os.rename(new_file_dirpath, filename + new_ext)
                      
        except OSError as e:
              print(e)
              logger.error(e)
              sys.exit(1)

def main():
    args = parser_args
    inf_sub: Dict[str, Any] = {}

    def guess_search(search:str):
        """ Parse search parameter. """

        # Custom configuration
        options:Dict[str, Any] = {
            'single_value': True,
            'excludes': ['country', 'language', 'audio_codec', 'other'],
            'output_input_string': True,
            'name_only': True
        }
        properties = ('type','title','season','episode','year')
        season = True if args.Season else False
        info = VideoMetadataExtractor.extract_specific(search, *properties, options=options)

        def _clean_search(search_param:str):
            """Remove special chars for `search_param`"""
            for i in [".", "-", "*", ":", ";", ","]:
                search_param = search_param.replace(i, " ")
            return search_param            

        try:

            if info["type"] == "episode":
                number = f"s{info['season']:02}e{info['episode']:02}" if all(i is not None for i in [info['season'], info['episode'], info['title']]) else ""
                
                if ( args.Season and all(i is not None for i in [ info['title'], info['season'] ]) )\
                    or all( i is not None for i in [info['season'], info['title']] ) and info['episode'] is None:
                    number = f"s{info['season']:02}"
                    season = True if number else season
            else:
                number = f"({info['year']})" if all(i is not None for i in [info['year'], info['title']]) else  ""

            if (args.title):
                title = f"{args.title}"
            else:
                if info["type"] == "movie":
                    title = f"{info['title'] if info['title'] is not None else _clean_search(search)}"
                else:
                    if all( i is not None for i in [ info["year"], info['title'] ] ):
                        title = f"{info['title']} ({info['year']})"
                    else:
                        title = f"{info['title']}" if all(i is not None for i in [ info['title'], info['season'] ])\
                                else _clean_search(search)
            inf_sub:Dict[str, Any] = {
                'type': info["type"],
                'season' : season,
                'number' : f"{number}"
            }

        except Exception as e:
            error = e.__class__.__name__
            msg = e.__str__()
            logger.debug(f"Failed to parse search argument: {search} {error}: {msg}")
            console.print(f":no_entry: [bold red]" + gl("Failed_to_parse_search_argument") + "[/] ", search, emoji=True)
            console.print(f"[red]{error}[/]: {msg}",emoji=True)
            sys.exit(1)

        return title, number, inf_sub

    def _exists_sub(filepath:str) -> bool:
        """ Check if exists the sub file"""
        exists_sub = False
        sub_file = os.path.splitext(filepath)[0]
        for ext in sub_extensions:
            if os.path.isfile(sub_file + ext):
                if args.force:
                  os.remove(sub_file + ext)
                else:
                    exists_sub = True
                    break
        if exists_sub:
            logger.debug(f'Subtitle already exits use -f for force downloading')
            if not args.quiet:
                console.print(":no_entry: [bold red]" + gl("Subtitle_already_exists") + "[/]", emoji=True)
        return exists_sub
    
    if not os.path.isdir(args.search):
        if not args.path:
            if _exists_sub(str(args.search)):
                sys.exit(1)
        try:
            search = f"{os.path.basename(args.search)}"
            title, number, inf_sub = guess_search(search)
            
            subid = get_subtitle_id(
                title, number, inf_sub)
        except NoResultsError as e:
            logger.error(str(e))
            sys.exit(1)
            
        if (subid):
            if args.path:
                topath = f'{args.path}'
                get_subtitle(subid, topath)
            elif os.path.isfile(args.search):
                filepath = os.path.join(os.getcwd(), str(args.search))
                topath = os.path.dirname(filepath)
                with subtitle_renamer(filepath, inf_sub):
                    get_subtitle(subid, topath)
            else:
                topath = os.getcwd()
                get_subtitle(subid, topath)

    elif os.path.exists(args.search):
      cursor = FindFiles(f'{args.search}', with_extension=_extensions)
      list_files:list[str] = cursor.findFiles()
      if not list_files:
            logger.debug(f'Not files to search in: {args.search}')
            if not args.quiet:
                console.print(":no_entry:[bold red] " + gl("Not_files_to_search_in") + "[/]", f'{args.search}', emoji=True)
            sys.exit(1)
    
      for filepath in list_files:
        # skip if a subtitle for this file exists
        if _exists_sub(filepath):
            continue

        filename = f'{os.path.basename(filepath)}'
        metadata:Metadata = extract_meta_data(filename, args.kword, True)
        
        try:
            title, number, inf_sub = guess_search(filename)

            subid = get_subtitle_id(
                title, number, inf_sub, metadata)

        except NoResultsError as e:
            logger.error(str(e))
            sys.exit(1)
        
        if not args.path:
            topath = f'{os.path.dirname(filepath)}'
        else:
            topath = f'{args.path}'

        if (subid):
            with subtitle_renamer(filepath, inf_sub):
                get_subtitle(subid, topath)

if __name__ == '__main__':
    main()



# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import re
import sys
import time
import json
import shutil
import signal
import certifi
import urllib3
import tempfile
import html2text
import urllib3.util
import typing
from zipfile import ZipFile, is_zipfile, ZipInfo
from rarfile import RarFile, is_rarfile, RarInfo # type: ignore
from sdx_dl.sdxparser import logger, args as parser_args
from sdx_dl.sdxclasses import HTML2BBCode, NoResultsError, GenerateUserAgent, VideoMetadataExtractor 
from json import JSONDecodeError
from urllib3.exceptions import HTTPError
from bs4 import BeautifulSoup
from typing import Dict, Any, NamedTuple, NewType
from itertools import chain
from datetime import datetime, timedelta
from readchar import readkey, key
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl

from rich import box
from rich.layout import Layout
from rich.console import Group
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.align import Align
from rich.live import Live
from rich.prompt import IntPrompt

__all__ = [
"get_imdb_search", "get_aadata", "convert_date", "get_filtered_results", "sort_results", "get_selected_subtitle_id",
"HTTPErrorsMessageException", "clean_screen", "paginate", "extract_subtitles", "sub_extensions", "Metadata", "metadata",
"SUBDIVX_DOWNLOAD_PAGE", "HTTPError", "headers", "conn"
]

args = parser_args

#obtained from https://flexget.com/Plugins/quality#qualities
_audio = ('dts-hd', 'dts', 'dd5.1', 'ddp5.1','atmos', 'truehd', 'aac', 'opus', 'flac', 'dolby')

sub_extensions = ['.srt', '.ssa', '.ass', '.sub']

_compressed_extensions = ['.zip', '.rar']

SUBDIVX_SEARCH_URL = 'https://www.subdivx.com/inc/ajax.php'

SUBDIVX_DOWNLOAD_PAGE = 'https://www.subdivx.com/'

Metadata = NamedTuple(
    'Metadata',
    [('keywords', list[str]),
    ('quality', list[str]),
    ('codec', list[str]),
    ('audio', list[str]),
    ('hasdata', bool)]
)

DataConn = NamedTuple('DataConn', [('cookie', str),('token', str), ('search', str)])

listDict = NewType('listDict', list[Dict[str,Any]])

signal.signal(signal.SIGINT, lambda _, __: sys.exit(0))

def clean_screen() -> None:
    """Clean the screen"""
    os.system('clear' if os.name != 'nt' else 'cls')

# Configure connections
ua = GenerateUserAgent.random_browser()
headers={"user-agent" : ua}

if args.proxy:
    proxie = f"{args.proxy}"
    if not (any(p in proxie for p in ["http", "https"])):
        proxie = "http://" + proxie
    conn = urllib3.ProxyManager(proxie, headers=headers, cert_reqs="CERT_REQUIRED", ca_certs=certifi.where(),  retries=2, timeout=15)
else:
    conn = urllib3.PoolManager(headers=headers, cert_reqs="CERT_REQUIRED", ca_certs=certifi.where(), retries=2, timeout=15)

# Network connections Errors
def HTTPErrorsMessageException(e: HTTPError) -> None:
    """ Manage HTTP Network connection Errors Exceptions message:
        * Log HTTP Network connection Error message
        * Print HTTP Network connection error message.
    """
    error_class = e.__cause__.__class__.__name__
    error_msg = gl(error_class)
    msg = "[bold yellow]" + error_class + ":[/] " + error_msg if error_class != error_msg else error_class

    if not args.quiet:clean_screen()
    console.print(":no_entry: [bold red]" + gl("Some_Network_Connection_Error_occurred") + "[/]: " + msg, new_line_start=True, emoji=True)
    logger.debug(f'Some Network Connection Error occurred: {e.__cause__.__str__()}')

### Setting data connection ###
class DataConnection:
    """
    Class for manage the connection data

    Attributes:
        _sdx_dc_path (str): Path to data connection file.
        __cookie (str): The cookie value for connections.
        __token (str):  The token for search requests.
        __search (str): Search field suffix.
        __data (DataConn): The entire data connection attributes.
        :type DataConn: NamedTuple
    """
    def __init__(self) -> None:
        self._sdx_dc_path = os.path.join(tempfile.gettempdir(), 'sdx_data_connection')
        self.__cookie, self.__token, self.__search = self._get_connection_data()
        self.__data = DataConn(self.__cookie, self.__token, self.__search)
    
    @property
    def cookie(self) -> str: return self.__data.cookie

    @property
    def token(self) -> str: return self.__data.token

    @property
    def search(self) -> str: return self.__data.search

    def _get_connection_data(self):
        """Return data connection"""
        sdx_data_connection = self._load_data_connection()

        if sdx_data_connection.count(";") != 2 or sdx_data_connection.count("=") != 1:
            logger.debug(f'Getting data connection')
            cookie, token, f_search = self._retrieve_data_connection()
        else:
            cookie, token, f_search = sdx_data_connection.split(";", maxsplit=2)
            logger.debug(f'Loaded data connection')
        
        return cookie, token, f_search

    def _exp_data_connection(self) -> bool:
        """Compare modified time and return `True` if is expired."""
        if not os.path.exists(self._sdx_dc_path): return True
        
        # Get data connection modified time and convert it to datetime
        csdx_ti_m = datetime.fromtimestamp(os.path.getmtime(self._sdx_dc_path))
        delta_csdx = datetime.now() - csdx_ti_m
        exp_c_time = timedelta(hours=2)

        return delta_csdx > exp_c_time

    def _retrieve_data_connection(self):
        """ Retrieve, save and return data connection."""
        cookie_sdx, _f_token, _f_search = f'', f'', f''

        try:
            sdx_request = conn.request('GET', SUBDIVX_DOWNLOAD_PAGE, timeout=10)
            cookie_sdx = f"{sdx_request.headers.get('Set-Cookie')}".split(';')[0]
            _vdata = BeautifulSoup(sdx_request.data.decode(), 'lxml')
            _f_search = str(_vdata('div', id="vs")[0].text.replace("v", "").replace(".", ""))
            _f_tk = SUBDIVX_SEARCH_URL[:-8] + 'gt.php?gt=1'
            _r_ftoken = conn.request('GET', _f_tk, headers={"Cookie":cookie_sdx},preload_content=False).data
            _f_token = f"{json.loads(_r_ftoken)['token']}"
        except Exception as e:
            if isinstance(e, (HTTPError)):
                HTTPErrorsMessageException(e)
            else:
                msg = e.__str__()
                console.print(":no_entry: [bold red]" + gl("Could_not_load_data_connection") + "[/]", emoji=True, new_line_start=True)
                logger.debug(f'Error: {e.__class__.__name__}: {msg}')
            sys.exit(1)
    
        with open(self._sdx_dc_path, 'w') as file:
            file.write(f'{cookie_sdx};{_f_token};{_f_search}')
            file.close()

        return cookie_sdx, _f_token, _f_search
        
    def _load_data_connection(self) -> str:
        """ Load stored sdx data connection return empty string if not exists or expired"""
        if not self._exp_data_connection():
            with open(self._sdx_dc_path, 'r') as filecookie:
                sdx_data_connection = filecookie.read()
        else:
            return f""

        return sdx_data_connection
    
    def reset_data_connection(self) -> None:
        """ Reset connection data """
        with open(self._sdx_dc_path, 'w') as file:
            file.write(f'')
            file.close()
    
conn_data = DataConnection()
headers['Cookie'] = conn_data.cookie

#### sdxlib utils ####
def extract_meta_data(search:str, kword:str, is_file:bool=False) -> Metadata:
    """
    Extract metadata from search based in matchs of keywords.

    The lists of keywords includen quality and codec for videos.
    """
    extractor = VideoMetadataExtractor()
    extracted_kwords = extractor.extract_specific(f"{search}", 'screen_size', 'video_codec',
                                                'release_group', 'source','streaming_service','other', options="-as")
    
    if ( all(x is None for x in extracted_kwords.values()) ):
        keywords = [x for x in f'{kword}'.split()[:4]] if kword else []
        return Metadata(keywords, [], [], [], bool(keywords))

    words = f""

    def clean_words(word:str):
        """clean words"""
        word = f'{word}'
        clean = [".", "-dl","-"]
        for i in clean:
            word = word.lower().replace(i, '')
        return word
    
    for k in ["release_group", "source", "streaming_service", "other"]:
        value = extracted_kwords[k]['raw'] if extracted_kwords[k] else None
        if (value):
            words += f"{value} " if k not in 'source' else f"{clean_words(value)} "
    
    words = words.strip()
    search = f'{search}'

    f = search.lower()[:-4] if is_file else search.lower()
    
    quality = [f"{extracted_kwords['screen_size']['value']}"] if extracted_kwords['screen_size'] else []
    codec = [clean_words(extracted_kwords['video_codec']['value'])] if extracted_kwords['video_codec'] else []
    audio = [o for o in _audio if o in f] or []
    keywords = [x for x in words.split()] if words else []

    #Split input keywords and add to the list   
    if (kword):
        keywords += f'{kword}'.split()[:4]
    
    # logger.debug(f'Keywords: {keywords} quality: {quality} codec: {codec} audio: {audio}')
    return Metadata(keywords, quality, codec, audio, True)

if not args.no_filter:
    metadata = extract_meta_data(args.search, args.kword, os.path.isfile(args.search))
else:
    metadata = Metadata([], [], [], [], False)

def sort_results(results_list:list[Dict[str,Any]], metadata:Metadata=metadata) -> list[Dict[str,Any]]:
    """
    Finding the `Metadata` (keywords, quality, codec, audio) in the description

    and order by `score`.
    
    ### Score:

    The scale of `scores`: `0.125` by `Metadata` and `0.5` for max `download`.

    `Score`-->`Metadata`: 0.125 --> 1 , 0.25 --> 2 , ... 0.50 --> 4
    """
    max_dl = max( [ int(x['descargas']) for x in results_list ] )
    results = listDict([])

    # compile patterns
    compile_keywords = re.compile(r'\b(?:' + '|'.join(map(re.escape, sorted(metadata.keywords, key=len, reverse=True))) + r')\b', flags=re.I)
    compile_quality = re.compile(r'\b(?:' + '|'.join(map(re.escape, sorted(metadata.quality, key=len, reverse=True))) + r')\b', flags=re.I)
    compile_codec = re.compile(r'\b(?:' + '|'.join(map(re.escape, sorted(metadata.codec, key=len, reverse=True))) + r')\b', flags=re.I)
    compile_audio = re.compile(r'\b(?:' + '|'.join(map(re.escape, sorted(metadata.audio, key=len, reverse=True))) + r')\b', flags=re.I)

    for subs_dict in results_list:
        description = f"{subs_dict['descripcion']}"
        score = 0
        meta = False
        
        if metadata.keywords and compile_keywords.search(description):
            score += .125
            meta = True

        if metadata.quality and compile_quality.search(description):
            score += .125
            meta = True

        if metadata.codec and compile_codec.search(description):
            score += .125
            meta = True

        if metadata.audio and compile_audio.search(description):
            score += .125
            meta = True

        if  max_dl == int(subs_dict['descargas']):
            score += .5

        subs_dict['score'] = score
        subs_dict['meta']  = True if meta else False
        results.append(subs_dict)

    results = sorted(results, key=lambda item: (item['score']), reverse=True)
      
    return results     

### Filters searchs functions ###
def match_text(title:str, number:str, inf_sub:Dict[str, Any], text:str):
  """Filter Search results with the best match possible"""

  # Setting Patterns
  special_char = ["`", "'", "´", ":", ".", "?"]
  for i in special_char:
      title = title.replace(i, '')
      text = text.replace(i, '')
  text = str(html2text.html2text(text)).strip()
  aka = "aka"
  search = f"{title} {number}"
  match_type = None
  rnumber = False
  raka = False
  
  # Setting searchs Patterns
  re_full_match = re.compile(rf"^{re.escape(search)}$", re.I)
  re_full_pattern = re.compile(rf"^{re.escape(title)}.*{number}.*$", re.I) if inf_sub['type'] == "movie"\
    else re.compile(rf"^{re.escape(title.split()[0])}.*{number}.*$", re.I)
  re_title_pattern = re.compile(rf"^{re.escape(title)}\b", re.I)

  # Perform searches
  r = True if re_full_match.search(text.strip()) else False
  match_type = 'full' if r else None

  if not r:
    r = True if re_full_pattern.search(text.strip()) else False
    match_type = 'pattern' if r else None 

  if not r :
    rtitle = True if re_title_pattern.search(text.strip()) else False
    for num in number.split(" "):
        if not inf_sub['season']:
           rnumber = True if re.search(rf"\b{num}\b", text, re.I) else False
        else:
           rnumber = True if re.search(rf"\b{num}.*\b", text, re.I) else False

    raka = True if re.search(rf"\b{aka}\b", text, re.I) else False

    if raka :
        r = True if rtitle and rnumber and raka else False
        match_type = 'partial' if r else None
    else:
        r = True if rtitle and rnumber else False
        match_type = 'partial' if r else None

  if not r:
    if all(re.search(rf"\b{word}\b", text, re.I) for word in search.split()) :
        r = True if rnumber and raka else False
        match_type = 'partial' if r else None

  if not r:
    if all(re.search(rf"\b{word}\b", text, re.I) for word in title.split()) :
        r = True if rnumber else False
        match_type = 'partial' if r else None

  if not r:
    match_type = 'any'

  return match_type 

def get_filtered_results (title:str, number:str, inf_sub:Dict[str, Any], list_Subs_Dicts:list[Dict[str, Any]]) -> list[Dict[str, Any]]:
    """Filter subtitles search for the best match results"""
    
    filtered_results = listDict([])
    lst_full = listDict([])
    lst_pattern = listDict([])
    lst_partial = listDict([])
    lst_any = listDict([])

    if inf_sub['type'] == "movie" and inf_sub['number'] == "":
        return list_Subs_Dicts

    for subs_dict in list_Subs_Dicts:
        mtype = match_text(title, number, inf_sub, subs_dict['titulo'])
        if mtype == 'full':
            lst_full.append(subs_dict)
        elif mtype == 'pattern':
            lst_pattern.append(subs_dict)
        elif mtype == 'partial':
            lst_partial.append(subs_dict)
        else:
            if mtype == 'any':
                lst_any.append(subs_dict)
    
    if inf_sub['season']:
        filtered_results = lst_full + lst_partial if len(lst_partial) !=0 else lst_full + lst_pattern
    
    if inf_sub['type'] == "episode":
        if (inf_sub['season']):
            if len(lst_full) != 0:
                filtered_results = lst_full + lst_pattern         
        else:
            if len(lst_full) != 0:
                filtered_results = lst_full
            elif len(lst_partial) != 0:
                filtered_results = lst_partial
            elif len(lst_pattern) != 0:
                filtered_results = lst_pattern
            else:
                filtered_results = lst_any
    
    
    if inf_sub['type'] == "movie":

        if len(lst_full) != 0 or len(lst_pattern) != 0:
            filtered_results = lst_full + lst_pattern
        elif len(lst_partial) != 0:
            filtered_results = lst_partial
        else:
            filtered_results = lst_any
    
    filtered_results = sorted(filtered_results, key=lambda item: item['id'], reverse=True)

    return filtered_results

### Filters searchs functions ###

@typing.no_type_check
def highlight_text(text:str, metadata:Metadata=metadata) -> str:
    """Highlight all `text`  matches  `metadata`"""
    
    # make a list of keywords and escaped it. Sort list for efficiency
    keywords = list(chain(*metadata[:4]))

    # compile a pattern
    matches_compile = re.compile(r'\b(?:' + '|'.join(map(re.escape, sorted(keywords, key=len, reverse=True))) + r')\b', flags=re.I)

    @typing.no_type_check
    def _highlight(matches):
        return "[white on green4]" + f'{matches.group(0)}' + "[default on default]" 

    highlighted = matches_compile.sub(_highlight, text) 
    
    return highlighted

def backoff_delay(backoff_factor:float = 2, attempts:int = 2) -> None:
    """ backoff algorithm: backoff_factor * (2 ** attempts)."""
    delay:float = backoff_factor * (2 ** attempts)
    time.sleep(delay)

def convert_datetime(string_datetime:str):
    """
       Convert ``string_datetime`` in a datetime obj then format it to "%d/%m/%Y %H:%M"

       Return ``--- --`` if not invalid datetime string.
    """

    try:
        date_obj = datetime.strptime(string_datetime, '%Y-%m-%d %H:%M:%S').date()
        time_obj = datetime.strptime(string_datetime, '%Y-%m-%d %H:%M:%S').time()
        date_time_str = datetime.combine(date_obj, time_obj).strftime('%d/%m/%Y %H:%M')

    except ValueError:
        return "--- --"
    
    return date_time_str

def convert_date(list_dict_subs:list[Dict[str, Any]]) -> list[Dict[str, Any]]:
    """   
    Convert to datetime Items ``fecha_subida``.
    """
    for dictionary in list_dict_subs:
        dictionary['fecha_subida'] = convert_datetime(str(dictionary['fecha_subida']))

    return list_dict_subs

def get_aadata(search:str) -> Any:
    """Get a json data with the ``search`` results."""
    json_aaData:Any = ''
    try:
        retries = urllib3.util.Retry(total=2, backoff_factor=1)
        fields:Dict[str, Any] = {'buscar' + conn_data.search: search,
        'filtros': '', 'tabla': 'resultados', 'token': conn_data.token
        }
        
        page = conn.request(
            'POST',
            SUBDIVX_SEARCH_URL,
            headers=headers,
            fields=fields,
            retries=retries
        ).data

        if not page :
            if not args.quiet: console.clear()
            console.print(":no_entry: [bold red]" + gl("Could_not_load_results_page") + "[/]", emoji=True, new_line_start=True)
            conn_data.reset_data_connection()
            logger.debug('Could not load results page')
            sys.exit(1)
        else :
            json_aaData = json.loads(page)
            if json_aaData['sEcho'] == "0":
                site_msg = str(json.loads(page)['mensaje'])
                logger.debug(f'Site message: {site_msg}')
                backoff_delay(backoff_factor=1.5)
                page = conn.request('POST', SUBDIVX_SEARCH_URL, headers=headers, fields=fields, retries=retries).data
                
                if page:
                    json_aaData = json.loads(page)
                    if json_aaData['sEcho'] == "0":
                        raise NoResultsError(f'Site message: {site_msg}')
                else:
                    sys.exit(1)
    
    except Exception as e:
        if isinstance(e, (HTTPError)):
            HTTPErrorsMessageException(e)
        else:
            msg = e.__str__()
            logger.debug(f'Error: {e.__class__.__name__}: {msg}')
            console.print(":no_entry: [bold red]" + gl("Could_not_load_results_page") + "[/]", emoji=True, new_line_start=True)
        sys.exit(1)
    
    return json_aaData

def make_layout() -> Layout:
    """Define the layout."""

    layout = Layout(name="results")

    layout.split_column(
        Layout(name="table")
    )
    
    return layout

def make_screen_layout() -> Layout:
    """Define a screen layout."""

    layout = Layout(name="screen")

    layout.split_column(
        Layout(name="subtitle"),
        Layout(name="description", size=8, ratio=1),
        Layout(name="caption")
    )
    layout["caption"].update(Align.center("Download:[[bold green]D[/]] Back:[[bold green]A[/]]",
                                          vertical="middle", style="italic bright_yellow"))

    return layout

def make_description_panel(description:str) -> Panel:
    """Define a description Panel."""

    descriptions = Table.grid(padding=1)
    descriptions.add_column()
    descriptions.add_row(description)
    descriptions_panel = Panel(
        Align.center(
            Group(Align.center(descriptions)), vertical = "middle"
        ),
        box = box.ROUNDED,
        title = "[bold yellow]Descripción:[/]",
        title_align = "left",
        subtitle = "[white on green4]Coincidencias[/] [italic bright_yellow]con los metadatos del archivo[/]",
        subtitle_align = "center",
        padding = 1 
    )

    return descriptions_panel

## Get Comments functions ##
def get_comments_data(subid:str):
    """Get comments Json data"""

    fields={'getComentarios': subid}
    try:
        page = conn.request('POST', SUBDIVX_SEARCH_URL, fields=fields, headers=headers).data
        json_comments = json.loads(page)
    except Exception as e:
        if isinstance(e, (HTTPError)):
            msg = e.__str__().split(":", maxsplit=1)[1].split("(")[0]
            logger.debug(f'Could not load comments ID:{subid}: Network Connection Error:{msg}')
        else:
            msg = e.__str__()
            logger.debug(f'Could not load comments ID:{subid}: Error: {e.__class__.__name__}: {msg}')
        return None

    return json_comments

def parse_list_comments(list_dict_comments:listDict) -> listDict:
    """ Parse comments :
       * Remove not used Items
       * Convert to datetime Items ``fecha_creacion``.
       * Convert ``nick`` to text
    """
    parser = html2text.HTML2Text()
    parser.ignore_images = True
    parser.ignore_links = True

    for dictionary in list_dict_comments:
        dictionary['fecha_creacion'] = convert_datetime(str(dictionary['fecha_creacion']))
        dictionary['nick'] = parser.handle(dictionary['nick']).strip()

    return list_dict_comments

def make_comments_table(title:str, results:Dict[str,Any], page:int, metadata:Metadata=metadata) -> Table:
    """Define a comments Table."""

    BG_STYLE = Style(color="white", bgcolor="gray0", bold=False)

    comment_table = Table(box=box.SIMPLE, title="\n" + title, caption="Prev.:[[bold green]\u2190[/]] Next:[[bold green]\u2192[/]] "\
                    "Back:[[bold green]A[/]] Download:[[bold green]D[/]] Metadata?:[green]Green[/]\n\n"\
                    "Pag.[bold white] " + str(page + 1) + "[/] of [bold white]" + str(results['pages_no']) + "[/] " \
                    "of [bold green]" + str(results['total']) + "[/] comment(s)",
                    show_header=True, header_style="bold yellow", title_style="bold green",
                    caption_style="italic bright_yellow", leading=0, show_lines=False, show_edge=False,show_footer=True)
    
    comment_table.add_column("#", justify="right", vertical="middle", style="bold green")
    comment_table.add_column("Comentarios", justify="left", vertical="middle", style="white")
    comment_table.add_column("Usuario", justify="center", vertical="middle")
    comment_table.add_column("Fecha", justify="center", vertical="middle")

    count = int(page * results['per_page'])
    rows: list[list[str]] = []
    items:list[str] = []
    
    for item in results['pages'][page]:
        try:
            comentario = html2text.html2text(item['comentario']).strip()
            if metadata.hasdata: comentario = highlight_text(comentario, metadata)
            usuario = str(item['nick'])
            fecha = str(item['fecha_creacion'])

            items = [str(count + 1), comentario, usuario, fecha]
            rows.append(items)
        except IndexError:
            pass
        count = count +1
    
    for row in rows:
        row[0] =  "[bold green]" + row[0] + "[/]"
        comment_table.add_row(*row, style = BG_STYLE )

    return comment_table

def not_comments(text:str) -> Panel:
    """Show Not Comments Panel"""

    not_comment_panel = Panel(
        Align.center(
            Group(Align.center(text,vertical='top')), vertical = "top"
        ),
        box = box.SIMPLE_HEAD,
        title = "[bold yellow]Comentarios[/]",
        subtitle ="Back:[[bold green]A[/]] Download:[[bold green]D[/]]",
        padding = 1,
        style="italic bright_yellow",
        height=5,
    )

    return not_comment_panel

### Show results and get subtitles ###

def generate_results(title:str, results:Dict[str, Any], page:int, selected:int, metadata:Metadata=metadata) -> Layout:
    """Generate Selectable results Table."""

    SELECTED = Style(color="white", bgcolor="gray35", bold=True)
    layout_results = make_layout() 

    table = Table(box=box.SIMPLE, title=">> " + f'{title}\n' +\
                "[italic]Pag.[bold white] " + f"{page + 1}" + "[/] of [bold white]" + f"{results['pages_no']}" +\
                "[/] of [bold green]" + f"{results['total']}" + "[/] result(s)[/]", 
                caption="\nDw:[bold green]\u2193[/] Up:[bold green]\u2191[/] Nx:[bold green]\u2192[/] Pv:[bold green]\u2190[/] "\
                "Dl:[bold green]ENTER[/] Descrip.:[bold green]D[/] Comments:[bold green]C[/] Exit:[bold green]S[/]\n" \
                "Date:[bold green]\u2193 PgDn[/] [bold green]\u2191 PgUp[/] Default:[bold green]F[/] Metadata?:[green]Green[/]",
                title_style="bold green", show_header=True, header_style="bold yellow", caption_style="bold bright_yellow",
                show_edge=False, pad_edge=False)
    
    table.add_column("#", justify="right", vertical="middle", style="bold green")
    table.add_column("Título", justify="left", vertical="middle", style="white", ratio=2)
    table.add_column("Descargas", justify="center", vertical="middle")
    table.add_column("Usuario", justify="center", vertical="middle")
    table.add_column("Fecha", justify="center", vertical="middle")

    count = int(page * results['per_page'])
    rows: list[list[str]] = []
    items:list[str] = []
 
    for item in results['pages'][page]:
        try:
            titulo = html2text.html2text(item['titulo']).strip()
            if metadata.hasdata: titulo = "[green]" + titulo + "[/]" if item['meta'] else titulo
            descargas = str(item['descargas'])
            usuario = str(item['nick'])
            fecha = str(item['fecha_subida'])

            items = [str(count + 1), titulo, descargas, usuario, fecha]
            rows.append(items)
        except IndexError:
            pass
        count = count +1
    
    for i, row in enumerate(rows):
        row[0] =  "[bold red]\u25cf[/]" + row[0] if i == selected else " " + row[0]
        table.add_row(*row, style=SELECTED if i == selected else "white")

    layout_results["table"].update(table)
    
    return layout_results

def paginate(items:list[Any], per_page:int) -> Dict[str, Any]:
    """ Paginate `items` in perpage lists 
    and return a `Dict` with:
     * Total items
     * Number of pages
     * Per page amount
     * List of pages.
    """
    pages = [items[i:i+per_page] for i in range(0, len(items), per_page)]
    results:Dict[str, Any] = {}
    results = {
        'total': len(items),
        'pages_no': len(pages),
        'per_page': per_page,
        'pages': pages
    }
    return results

def get_rows() -> int:
    """Get Terminal available rows"""
    lines = shutil.get_terminal_size().lines
    fixed_lines = lines - 10
    available_lines = fixed_lines if (fixed_lines > 0) else lines
    if args.nlines:
        num_lines = args.nlines
        available_lines = min(available_lines, num_lines)

    return available_lines

def get_comments_rows() -> int:
    """Get Terminal available rows for comments"""
    lines = shutil.get_terminal_size().lines
    fixed_lines = lines - 20
    available_lines = fixed_lines if (fixed_lines > 0) else lines
    if args.nlines:
        num_lines = args.nlines
        available_lines = min(available_lines, num_lines)

    return available_lines

def get_selected_subtitle_id(table_title:str, results:list[Dict[str,Any]], metadata:Metadata=metadata) -> str:
    """Show subtitles search results for obtain download id."""

    results_pages = paginate(results, get_rows())
    selected = 0
    page = 0
    res = f""

    try:
        with Live(
            generate_results (table_title, results_pages, page, selected, metadata),auto_refresh=False, screen=False, transient=True
        ) as live:
            while True:
                live.console.show_cursor(False)

                ch = readkey()
                if ch == key.UP:
                    selected = max(0, selected - 1)
                
                if ch == key.PAGE_UP:
                    results_pages = sorted(results, key=lambda item: (
                                    datetime.strptime(item['fecha_subida'],'%d/%m/%Y %H:%M')
                                    if item['fecha_subida'] != "--- --" else datetime.min
                                    ), reverse=False
                                )
                    results_pages = paginate(results_pages, get_rows())
                
                if ch == key.PAGE_DOWN:
                    results_pages = sorted(results, key=lambda item: (
                                    datetime.strptime(item['fecha_subida'],'%d/%m/%Y %H:%M')
                                    if item['fecha_subida'] != "--- --" else datetime.min
                                    ), reverse=True
                                )
                    results_pages = paginate(results_pages, get_rows())
                
                if ch in ["F", "f"]:
                      results_pages = paginate(results, get_rows())
                
                if ch == key.DOWN:
                    selected = min(len(results_pages['pages'][page]) - 1, selected + 1)

                if ch in ["D", "d"]:
                    description_selected = str(results_pages['pages'][page][selected]['descripcion'])
                    subtitle_selected =  results_pages['pages'][page][selected]['titulo']
                    parser = HTML2BBCode()
                    description = parser.html_to_bbcode(description_selected)
                    description = highlight_text(description, metadata) if metadata.hasdata else description

                    layout_description = make_screen_layout()
                    layout_description["description"].update(make_description_panel(description))
                    layout_description["subtitle"].update(Align.center(
                                html2text.html2text(subtitle_selected).strip(),
                                vertical="middle",
                                style="italic bold green"
                                ))

                    with console.screen(hide_cursor=True) as screen: 
                        while True:
                            screen.console.show_cursor(False)
                            screen.update(layout_description)

                            ch_exit = readkey()
                            if ch_exit in ["A", "a"]:
                                break

                            if ch_exit in ["D", "d"]:
                                res = f"{results_pages['pages'][page][selected]['id']}"
                                break
                                
                    if res != "": break
                
                if ch in ["C", "c"]:
                    cpage = 0
                    subtitle_selected =  results_pages['pages'][page][selected]['titulo']
                    subid = str(results_pages['pages'][page][selected]['id'])
                    layout_comments = make_layout()
                    title = html2text.html2text(subtitle_selected).strip()
                    show_comments = True if results_pages['pages'][page][selected]['comentarios'] != 0 else False
                    comment_msg = ":neutral_face: [bold red][i]¡No hay comentarios para este subtítulo![/]" if not show_comments else "" 
                    comments = {}

                    with console.screen(hide_cursor=True) as screen_comments:
                        if show_comments:
                            with console.status("[bold yellow][i]CARGANDO COMENTARIOS...[/]", spinner='aesthetic'):
                              aaData = get_comments_data(subid)
                            if aaData: 
                                comments = aaData['aaData']
                                comments = parse_list_comments(comments)
                                comments = paginate(comments, get_comments_rows())
                            
                            if not comments:
                                show_comments = False
                                comment_msg = ":neutral_face: [bold red][i]¡No se pudieron cargar los comentarios![/]"
                        
                        while True:
                            if show_comments :
                                layout_comments['table'].update(Align.center(
                                    Group(Align.center(make_comments_table(title, comments, cpage, metadata), vertical="top")), vertical='top'
                                    )
                                )
                            else :
                                layout_comments['table'].update(not_comments(comment_msg))
                            
                            screen_comments.console.show_cursor(False)
                            screen_comments.update(layout_comments)

                            ch_comment = readkey()
                            
                            if ch_comment in ["A", "a"]:
                                break
                            
                            if ch_comment == key.RIGHT :
                                cpage = min(comments["pages_no"] - 1, cpage + 1)

                            if ch_comment == key.LEFT :
                                cpage = max(0, cpage - 1)

                            if ch_comment in ["D", "d"]:
                                res = subid
                                break

                    if res != "": break

                if ch == key.RIGHT :
                    page = min(results_pages["pages_no"] - 1, page + 1)
                    selected = 0

                if ch == key.LEFT :
                    page = max(0, page - 1)
                    selected = 0

                if ch == key.ENTER:
                    res = f"{results_pages['pages'][page][selected]['id']}"
                    break

                if ch in ["S", "s"]:
                    res = -1
                    break
                live.update(generate_results(table_title, results_pages, page, selected, metadata), refresh=True)

    except KeyboardInterrupt:
        if not args.verbose:clean_screen()
        logger.debug('Interrupted by user')
        sys.exit(1)

    if (res == -1):
        if not args.verbose:clean_screen()
        logger.debug('Download Canceled')
        return f""
    
    # clean_screen()
    return res

### Extract Subtitles ###
@typing.no_type_check
def extract_subtitles(compressed_sub_file: ZipFile | RarFile, topath:str) -> None:
    """Extract ``compressed_sub_file`` from ``temp_file`` ``topath``."""
    # For portable Windows EXE
    if sys.platform == "win32":
        import rarfile #type: ignore

        @typing.no_type_check
        def resource_path(relative_path:str) -> str:
            """ Get absolute path to resource, works for dev and for PyInstaller """
            base_path:str = ""
            try:
                # PyInstaller creates a temp folder and stores path in _MEIPASS
                if hasattr(sys, '_MEIPASS'):
                    base_path = sys._MEIPASS
            except Exception:
                base_path = os.path.abspath(".")
            
            return os.path.join(base_path, relative_path)
            
        rarfile.UNRAR_TOOL = resource_path("UnRAR.exe")

    def _is_supported(sub:RarInfo | ZipInfo) -> bool:
        filename = f'{os.path.basename(str(sub.filename))}'
        return any(filename.endswith(ext) for ext in sub_extensions) and '__MACOSX' not in filename

    def _is_compressed(filename: str) -> bool:
        """Check if a file is a supported archive based on its extension."""
        return any(filename.endswith(ext) for ext in _compressed_extensions)
    
    @typing.no_type_check
    def _uncompress(source:Any, topath:str) -> None:
        """Decompress compressed file"""
        compressed = RarFile(source) if is_rarfile(source) else ZipFile(source) if is_zipfile(source) else None
        if compressed:
            for sub in compressed.infolist(): 
                if _is_compressed(sub.filename):
                    source = compressed.open(sub)
                    _uncompress(source, topath)
                else:
                    compressed.extract(sub, topath)
            compressed.close()
        else:
            logger.debug(f'Unsupported archive format')

    # In case of existence of various subtitles choose which to download
    if len(compressed_sub_file.infolist()) > 1 : 
        res:int = 0
        count:int = 0
        choices:list[str] = []
        choices.append(str(count))
        list_sub:list[str] = []

        for i in compressed_sub_file.infolist():
            if i.is_dir() or os.path.basename(i.filename).startswith("._"): # type: ignore
                continue
            i.filename = os.path.basename(i.filename)
            list_sub.append(f'{i.filename}')
        
        if not args.no_choose:
            clean_screen()
            table = Table(box=box.ROUNDED, title=">> Subtítulos disponibles:", title_style="bold green",show_header=True, 
                        header_style="bold yellow", show_lines=True, title_justify='center')
            table.add_column("#", justify="center", vertical="middle", style="bold green")
            table.add_column("Subtítulos", justify="center" , no_wrap=True)

            for i in list_sub:
                table.add_row(str(count + 1), i)
                count += 1
                choices.append(str(count))
        
            choices.append(str(count + 1))
            console.print(table)
            console.print("[bold green]>> [0] " + gl("Download_all"), new_line_start=True)
            console.print("[bold red]>> [" + str(count + 1) + "] " + gl("Cancel_download"), new_line_start=True)

            try:
                res = IntPrompt.ask("[bold yellow]>> " + gl("Choose_a") + "[" + "[bold green]#" + "][bold yellow]." + gl("By_default") , 
                            show_choices=False, show_default=True, choices=choices, default=0)
            except KeyboardInterrupt:
                logger.debug('Interrupted by user')
                if not args.quiet:
                    console.print(":x: [bold red]I " + gl("Interrupted_by_user"), emoji=True, new_line_start=True)
                    time.sleep(0.4)
                    clean_screen()
                return
        
            if (res == count + 1):
                logger.debug('Canceled Download Subtitle') 
                if not args.quiet:
                    console.print(":x: [bold red] " + gl("Canceled_Download_Subtitle"), emoji=True, new_line_start=True)
                    time.sleep(0.4)
                    clean_screen()
                return

            clean_screen()

        logger.debug('Decompressing files')

        if res == 0:
            with compressed_sub_file as csf:
                for sub in csf.infolist():
                    if _is_supported(sub) and not sub.is_dir():
                        logger.debug(' '.join(['Decompressing subtitle:', sub.filename, 'to', topath]))
                        csf.extract(sub, topath)
                    elif _is_compressed(sub.filename):
                        with csf.open(sub) as source:
                            _uncompress(source, topath)
                        logger.debug(' '.join(['Decompressed file:', sub.filename, 'to', topath]))         
            compressed_sub_file.close()
        else:
            selected = f'{list_sub[res - 1]}'
            with compressed_sub_file as csf:
                for sub in csf.infolist():
                    if selected == os.path.basename(sub.filename):
                        if _is_supported(sub):
                            logger.debug(' '.join(['Decompressing subtitle:', selected, 'to', topath]))
                            csf.extract(sub, topath)
                        elif _is_compressed(sub.filename):
                            with csf.open(sub) as source:
                                _uncompress(source, topath=topath)
                            logger.debug(' '.join(['Decompressed file:', sub.filename, 'to', topath]))
                        break
            compressed_sub_file.close()

        logger.debug(f"Done extract subtitles!")

        if not args.quiet:
            clean_screen()
            console.print(gl("Done_download_subtitles"), emoji=True, new_line_start=True)
    else:
        for name in compressed_sub_file.infolist():
            # don't unzip stub __MACOSX folders
            if _is_supported(name):
                logger.debug(' '.join(['Decompressing subtitle:', name.filename, 'to', topath]))
                compressed_sub_file.extract(name, topath)
            elif _is_compressed(name.filename):
                with compressed_sub_file.open(name) as source:
                    _uncompress(source, topath=topath)
                logger.debug(' '.join(['Decompressed file:', name.filename, 'to', topath]))
        compressed_sub_file.close()

        logger.debug(f"Done extract subtitle!")
        if not args.quiet: console.print(gl("Done_download_subtitle"), emoji=True, new_line_start=True)

### Search IMDB ###

def get_imdb_search(title:str, number:str, inf_sub:Dict[str, Any]):
    """Get the IMDB ``id`` or ``title`` for search subtitles"""
    from sdx_dl.sdximdb import IMDB
    try:
        imdb = IMDB()
        if args.proxy:
            proxies = {'http': proxie, 'https': proxie}
            imdb.session.proxies.update(proxies)
            imdb.session.verify = False
        
        year = int(number[1:5]) if (inf_sub['type']  == "movie") and (number != "") else None

        if inf_sub['type'] == "movie":
            res = imdb.get_by_name(title, year, tv=False) if year else imdb.search(title, tv=False) # type: ignore
        else:
            res = imdb.search(title, tv=True) # type: ignore
    except Exception:
        pass
        return None
    
    try:
        results = json.loads(res) if year else json.loads(res)['results']
    except JSONDecodeError as e:
        msg = e.__str__()
        logger.debug(f'Could not decode json results: Error JSONDecodeError:"{msg}"')
        if not args.quiet: console.print(":no_entry: [bold red] " + gl("Some_error_retrieving_from_IMDB") + "[/]: " + msg,\
                                        new_line_start=True, emoji=True)
        return None
    
    if not results:
        return None
    else:
        if "result_count" in results and not results['results']:
            return None

    if year:
        search = f"{results['id']}" if inf_sub['type'] == "movie" else f"{results['name']} {number}"
        return search
    else:
        search = make_IMDB_table(title, results, inf_sub['type'])
        if inf_sub['type'] == "movie":
            return search
        else:
            return f'{search} {number}' if search else None

def make_IMDB_table(title:str, results:list[Any], type:str):
    """Define a IMDB Table."""
    count = 0
    choices:list[str] = []
    choices.append(str(count))

    BG_STYLE = Style(color="white", bgcolor="gray0", bold=False)

    imdb_table = Table(box=box.SIMPLE_HEAD, title="\n Resultados de IMDB para: " + title, caption="[italic bright_yellow]"\
                    "Seleccione un resultado o enter para cancelar[/]\n",
                    show_header=True, header_style="bold yellow", title_style="bold green",
                    caption_style="bold bright_yellow", leading=1, show_lines=True)
    
    imdb_table.add_column("#", justify="right", vertical="middle", style="bold green")
    imdb_table.add_column("Título + url", justify="left", vertical="middle", style="white")
    imdb_table.add_column("IMDB", justify="center", vertical="middle")
    imdb_table.add_column("Tipo", justify="center", vertical="middle")

    rows:list[list[str]] = []
 
    for item in results:
        try:
            titulo = str(html2text.html2text(item['name'])).strip() + " ("+ str(item['year'])+ ")\n" + str(item['url'])
            imdb = str(item['id'])
            tipo = str(item['type'])

            items = [str(count + 1), titulo, imdb, tipo]
            choices.append(str(count + 1))
            rows.append(items)
        except IndexError:
            pass
        count = count +1
    
    for row in rows:
        row[0] =  "[bold green]" + row[0] + "[/]"
        imdb_table.add_row(*row, style = BG_STYLE )
    
    console.print(imdb_table)
    console.print("[bold green]>> [0] Cancelar selección\n\r", new_line_start=True)
    
    res = IntPrompt.ask("[bold yellow]>> Elija un [" + "[bold green]#" + "][bold yellow]. Por defecto:", 
                    show_choices=False, show_default=True, choices=choices, default=0)
  
    search = f"{results[res-1]['id']}" if type == "movie" else f"{results[res-1]['name']}"

    return search if res else None



# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import tempfile
import argparse
import logging
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import set_locale, gl
from sdx_dl.sdxclasses import ChkVersionAction, ConfigManager, ViewConfigAction, SaveConfigAction, SetConfigAction, ResetConfigAction, validate_proxy
from importlib.metadata import version
from rich.logging import RichHandler
from rich.traceback import install
install(show_locals=True)

__all__ = ["args", "logger"]

def create_parser():
    parser = argparse.ArgumentParser(prog='sdx-dl',
    formatter_class=argparse.RawTextHelpFormatter,
    usage="sdx-dl [options] search",
    description='A cli tool for download subtitle from https://www.subdivx.com with the better possible matching results.',
    epilog='Project issues:https://github.com/Spheres-cu/subdx-dl/issues\n\
            \nUsage examples:https://github.com/Spheres-cu/subdx-dl#examples\n\n'
    )

    parser.add_argument('search', type=str,help="file, directory or movie/series title or IMDB Id to retrieve subtitles")

    parser.add_argument('--quiet', '-q', action='store_true',default=False, help="No verbose mode")
    parser.add_argument('--verbose', '-v', action='store_true',default=False, help="Be in verbose mode")
    parser.add_argument('--force', '-f', action='store_true',default=False, help="override existing file")
    parser.add_argument('--no-choose', '-nc', action='store_true', default=False, help="No Choose sub manually")
    parser.add_argument('--no-filter', '-nf', action='store_true',default=False, help="Do not filter search results")
    parser.add_argument('--nlines', '-nl', type=int, choices=[5, 10, 15, 20], default=False, nargs='?', const=10,
                        help="Show nl(5,10,15,20) availables records per screen. Default 10 records.", metavar="")
    parser.add_argument('--lang', '-l', type=str, choices=["es", "en"], default=False, nargs='?', const="es",
                         help="Show messages in language es or en", metavar="")
    parser.add_argument('--version', '-V', action='version', version=f'subdx-dl {version("subdx-dl")}', help="Show program version")
    parser.add_argument('--check-version', '-cv', action=ChkVersionAction, help="Check for new version")
                        
    ## Download opts group
    download_opts = parser.add_argument_group('Download')
    download_opts.add_argument('--path', '-p', type=str, help="Path to download subtitles")
    download_opts.add_argument('--proxy', '-x', type=str, help="Set a http(s) proxy(x) connection", metavar="x")

    ## Search opts group
    search_opts = parser.add_argument_group('Search by')
    search_opts.add_argument('--Season', '-S', action='store_true',default=False, help="Search by Season")
    search_opts.add_argument('--kword','-k',type=str,help="Add keywords to search among subtitles descriptions", metavar="kw")
    search_opts.add_argument('--title','-t',type=str,help="Set the title to search", metavar="t")
    search_opts.add_argument('--imdb', '-i', action='store_true',default=False, help="Search first for the IMDB id or title")

    ## Config opts group
    config_opts = parser.add_argument_group('Config').add_mutually_exclusive_group()
    config_opts.add_argument('--view-config', '-vc', action=ViewConfigAction, help="View config file")
    config_opts.add_argument('--save-config', '-sc', action=SaveConfigAction, help="Save options to config file")
    config_opts.add_argument('--load-config', '-lc',action='store_true', default=False, help="Load config file options")
    
    config_opts.add_argument('--config', '-c',
                            action=SetConfigAction,
                            choices=["quiet", "verbose", "force", "no_choose", "no_filter", "nlines", "path", "proxy", "Season", "imdb", "lang"],
                            nargs='?',metavar="o",help="Save an option[o] to config file")
    config_opts.add_argument('--reset', '-r',
                            action=ResetConfigAction,
                            choices=["quiet", "verbose", "force", "no_choose", "no_filter", "nlines", "path", "proxy", "Season", "imdb", "lang"],
                            metavar="o",help="Reset an option[o] in the config file")
 
    return parser

def create_logger(level:str = "DEBUG", verbose:bool=False):

    # Setting logger
    levels = ["CRITICAL", "ERROR", "WARNING","INFO", "DEBUG"]
    LOGGER_LEVEL = levels[4]
    LOGGER_FORMATTER_LONG = logging.Formatter('%(asctime)-12s %(levelname)-6s %(message)s', '%Y-%m-%d %H:%M:%S')
    LOGGER_FORMATTER_SHORT = logging.Formatter(fmt='%(message)s', datefmt="[%X]")

    level = level if level in levels else LOGGER_LEVEL
    temp_log_dir = tempfile.gettempdir()
    file_log = os.path.join(temp_log_dir, 'subdx-dl.log') 

    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(level)
    
    if not verbose:
        logfile = logging.FileHandler(file_log, mode='w', encoding='utf-8')
        logfile.setFormatter(LOGGER_FORMATTER_LONG)
        logfile.setLevel(level)
        logger.addHandler(logfile)
    else:
        console = RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)
        console.setFormatter(LOGGER_FORMATTER_SHORT)
        console.setLevel(level)
        logger.addHandler(console)

parser = create_parser()
args = parser.parse_args()

if args.load_config:
    config = ConfigManager()
    if config.exists and config.hasconfig:
        copied_args = args.__dict__.copy()
        new_args = config.merge_config(copied_args)

        for k, v in new_args.items():
            args.__setattr__(k, v)

if args.verbose: args.__setattr__("quiet", True)
create_logger(verbose=args.verbose)

if args.load_config:
     logger.debug("Config loaded!")

if args.lang:
    set_locale(args.lang)
else:
    config = ConfigManager()
    if config.hasconfig and 'lang' in config.config:
        set_locale(config.config['lang'])

if args.path:
    if not (os.path.isdir(args.path) and os.access(args.path, os.W_OK)):
        if args.quiet:
            logger.debug(f'Directory {args.path} do not exists')
        else:
            console.print(":no_entry:[bold red] " + gl("Directory") + "[/]" + f"{args.path}" + "[bold red] " + gl("Directory_not_exists") + "[/]",
                        new_line_start=True, emoji=True) 
        sys.exit(1)

if (args.proxy and not validate_proxy(args.proxy)):
    if args.quiet:
        logger.debug(f'Incorrect proxy setting. Only http, https or IP/domain:PORT is accepted')
    else:
        console.print(":no_entry:[bold red] " + gl("Incorrect_proxy_setting").split('.')[0] + ":[yellow] " + f"{args.proxy}" + "[/]",
                    new_line_start=True, emoji=True)
    sys.exit(1)



# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl, set_locale

### Config Settings imports ###
import os
import json
import typing
from typing import Optional, List, Dict, Any
from pathlib import Path

### Metadata video extractor imports ###
from guessit import guessit, jsonutils # type: ignore

### Check version imports ###
import re
import argparse
import urllib3
import certifi
from bs4 import BeautifulSoup
from importlib.metadata import version
from urllib3.exceptions import HTTPError

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

####  HTML2BBCode imports ###
from bs4 import Tag
from pygments.lexers import guess_lexer

####  Generate user agents imports ###
import sys
import random

__all__ = [
    "HTML2BBCode",
    "NoResultsError",
    "GenerateUserAgent",
    "VideoMetadataExtractor",
    "validate_proxy",
    "ChkVersionAction",
    "ConfigManager",
    "ViewConfigAction",
    "SaveConfigAction",
    "SetConfigAction",
    "ResetConfigAction",
    "FindFiles"
]

####  HTML2BBCode class ###
class HTML2BBCode:
    """
    HTML to BBCode converter
    """
    @staticmethod
    def guess_language(code: str) -> str:
        """
        Guesses the programming language from a string. This is not accurate for short strings. 'pygments' is used to
        guess code here because many other language detectors cannot detect the language properly.
        """
        guessed = guess_lexer(code).name
        if guessed == "Text only":
            return ""
        return guessed.lower()

    @classmethod
    def _html_to_bbcode(cls, tag:Tag) -> str:
        result = f""
        if hasattr(tag, 'children'):
            for child in tag.children:
                if isinstance(child, str):
                    result += child
                elif isinstance(child, Tag):
                    result += cls._html_to_bbcode(child) 

        if tag.name == "span":
            if "class" in list(tag.attrs.keys()):
                tag_type = tag["class"][0]
                if tag_type == "bb-bold":  # bold
                    result = f"[bold green]{result}[/bold green]"
                elif tag_type == "bb-italic":  # italic
                    result = f"[italic bright_yellow]{result}[/italic bright_yellow]"
                elif tag_type == "bb-underline":  # underline
                    result = f"[u]{result}[/u]"
                elif tag_type == "bb-strikethrough":  # strikethrough
                    result = f"[s]{result}[/s]"
                elif tag_type == "bb-big":  # big
                    result = f"[big]{result}[/big]"
                elif tag_type == "bb-small":  # small
                    result = f"[small]{result}[/small]"
            elif "style" in list(tag.attrs.keys()):
                if "color" in tag["style"]:  # color
                    tag_color = f'{tag.get("style")}'
                    result = f'[color={tag_color.replace("color:", "")}]{result}[/color]'
        elif tag.name == "br":  # line break
            result = "\n"
        elif tag.name == "a":  # url
            result = f"[link={tag['href']}]{result}[/link]"
        elif tag.name == "b":  # bold
            result = f"[bold green]{result}[/bold green]"
        elif tag.name == "i":  # bold
            result = f"[italic bright_yellow]{result}[/italic bright_yellow]" # italic
        elif tag.name == "ul":  # unordered list
            result = f"[list]\n{result}[/list]"
        elif tag.name == "li":  # list item
            result = f"[*] {result}"
        elif tag.name == "ol":  # ordered list
            result = f"[list=1]\n{result}[/list]"
        elif tag.name == "blockquote":  # quote
            quote_author = tag.find("p", {"class": "bb-quote-author"})
            if quote_author is not None:
                quote_author_name = str(quote_author.get_text()).replace(" wrote:", "")
                quote_author = "=" + quote_author_name
            else:
                quote_author = ""
            result = f"[quote{quote_author}]{result}[/quote]"
        elif tag.name == "p":
            if "class" in list(tag.attrs.keys()):
                if "bb-quote-author" in tag["class"]:
                    result = ""
            else:
                result = f"[p]{result}[/p]"  # p
        elif tag.name == "div":
            if "class" in list(tag.attrs.keys()):
                if "code" in tag["class"]:  # code
                    language = cls.guess_language(tag.get_text())
                    if not language:
                        language = ""
                    else:
                        language = "=" + language
                    result = f"[code{language}]{result}[/code]"
                # This converter cannot convert scratchblocks html to scratchblocks bbcode
                elif "scratchblocks" in tag["class"]:  # scratchblocks
                    result = f"[scratchblocks]{tag.get_text()}[/scratchblocks]"
            elif "style" in list(tag.attrs.keys()):
                if "text-align:center;" in tag["style"]:  # center
                    result = f"[center]{result}[/center]"
        return result

    @classmethod
    def html_to_bbcode(cls, html: str) -> str:
        try:
            soup = BeautifulSoup(html, "lxml")  # lxml is the fastest
            return cls._html_to_bbcode(soup)
        except Exception:
            pass
            return html

####  Utils Classes ###
class NoResultsError(Exception):
    pass

### Generate a user agent class ###
class GenerateUserAgent:
    """
    Class containing methods for generating user agents.
    """

    @staticmethod
    def _token() -> str:
        return "Mozilla/5.0"
    
    @staticmethod
    def _platform() -> str:
        _WINDOWS_PREFIX: str = "Windows NT 10.0; Win64; x64"
        _MAC_PREFIX: str = "Macintosh; Intel Mac OS X"
        _LINUX_PREFIX: str = "X11; Ubuntu; Linux x86_64"

        if sys.platform == "win32":
            # Windows
            platform = _WINDOWS_PREFIX
        elif sys.platform == "darwin":
            # macOS
            platform = _MAC_PREFIX
        else:
            # Linux and other UNIX-like systems
            platform = _LINUX_PREFIX
        return f'{platform}'

    @classmethod
    def firefox(cls) -> list[str]:
        """Generate a list of common firefox user agents

        Returns:
            list[str]: The list of common firefox user agents
        """
        return [f"{cls._token()} ({cls._platform()}; rv:{version}.0) Gecko/20100101 Firefox/{version}.0" for version in range(120, 138)]

    @classmethod
    def chrome(cls) -> list[str]:
        """Generate a list of common chrome user agents

        Returns:
            list[str]: The list of common chrome user agents
        """
        return [f"{cls._token()} ({cls._platform()}) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version}.0.0.0 Safari/537.36" for version in range(120, 135)]

    @classmethod
    def opera(cls) -> list[str]:
        """Generate a list of common opera user agents

        Returns:
            list[str]: The list of common opera user agents
        """
        return [f"{cls._token()} ({cls._platform()}) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version}.0.0.0 Safari/537.36 OPR/{opr}.0.0.0"
            for version in range(120, 135, 5) for opr in range(103, 118, 5)]

    @classmethod
    def safari(cls) -> list[str]:
        """Generate a list of common safari user agents

        Returns:
            list[str]: The list of common safari user agents
        """
        if sys.platform == "darwin":
            return [f"{cls._token()} ({cls._platform()} 14_7_5) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/{major}.{minor} Safari/605.1.15" for major, minors in [(16, range(5, 7)), (17, range(0, 7))] for minor in minors]
        else:
            return []

    @classmethod
    def safari_mobile(cls) -> list[str]:
        """Generate a list of common mobile safari user agents

        Returns:
            list[str]: The list of common safari mobile user agents
        """
        if sys.platform == "darwin":
            return [f"{cls._token()} (iPhone; CPU iPhone OS {major}_{minor} like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/{major}.{minor} Mobile/15E148 Safari/604.1" for major, minors in [(16, range(5, 8)), (17, range(0, 7))] for minor in minors]
        else:
            return []

    @staticmethod
    def generate_all() -> list[str]:
        """Convenience method, Generate common user agents for all supported browsers

        Returns:
            list[str]: The list of common user agents for all supported browsers in GenerateUserAgent.
        """
        if sys.platform == "darwin":
            return GenerateUserAgent.safari() + GenerateUserAgent.safari_mobile() + GenerateUserAgent.opera()
        else:
            return GenerateUserAgent.firefox() + GenerateUserAgent.chrome() + GenerateUserAgent.opera()
        
    @staticmethod
    def generate_random() -> list[str]:
        """Convenience method, Generate random user agents for all supported browsers

        Returns:
            list[str]: The list of random user agents for all supported browsers in GenerateUserAgent.
        """
        if sys.platform == "darwin":
            return random.choice([GenerateUserAgent.safari() + GenerateUserAgent.safari_mobile() + GenerateUserAgent.opera()])
        else:
            return random.choice([GenerateUserAgent.firefox(), GenerateUserAgent.chrome(), GenerateUserAgent.opera()])
        
    @staticmethod
    def random_browser() -> str:
        """Convenience method, Generate a random user agents for one supported browser

        Returns:
            str: With the random user agents for one supported browser in GenerateUserAgent.
        """
        if sys.platform == "darwin":
            browser = random.choice([GenerateUserAgent.safari() + GenerateUserAgent.safari_mobile() + GenerateUserAgent.opera()])
            return random.choice(browser)
        else:
            browser = random.choice([GenerateUserAgent.firefox(), GenerateUserAgent.chrome(), GenerateUserAgent.opera()])
            return random.choice(browser)

### validate proxy settings ###

def validate_proxy(proxy_str:str) -> bool:
    """
    Validation with IP address or domain and port.
    """

    ip_pattern = r'^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'
    host_pattern = r'^[a-zA-Z0-9][a-zA-Z0-9.-]*[a-zA-Z0-9]$'
    
    match = re.match(r'^(?:(https|http)://)?(?:([^:@]+):([^:@]+)@)?([^:@/]+)(?::(\d+))?$', proxy_str)
    
    if not match:
        return False
        
    protocol, _, _, host, port = match.groups() 
    
    if not (re.match(ip_pattern, host) or re.match(host_pattern, host)):
        return False
        
    if (port is None) or (not (0 < int(port) <= 65535)):
        return False
    
    if protocol not in ["http", "https", None]:
        return False
    
    return True

### Check version ###
def ExceptionErrorMessage(e: Exception):
    """Parse ``Exception`` error message."""
    if isinstance(e, (HTTPError)):
        msg = e.__str__().split(":", maxsplit=1)[1].split("(")[0]
    else:
        msg = e.__str__()
    error_class = e.__class__.__name__
    console.print(gl("Error_occurred") + error_class + ":" + msg)
    sys.exit(1)

def _check_version(version:str, proxy:str):
    """Check for new version."""
    ua = GenerateUserAgent.random_browser()
    headers={"user-agent" : ua}
    msg = f''
    
    if (proxy):
        if not (any(p in proxy for p in ["http", "https"])):
            proxy = "http://" + proxy
        session = urllib3.ProxyManager(proxy, headers=headers, cert_reqs="CERT_REQUIRED", ca_certs=certifi.where(), retries=2, timeout=10)
    else:
        session = urllib3.PoolManager(headers=headers, cert_reqs="CERT_REQUIRED", ca_certs=certifi.where(), retries=2, timeout=10)
  
    def get_version_description(version:str):
        """Get new `version` description."""
        url = f"https://github.com/Spheres-cu/subdx-dl/releases/tag/{version}"
        
        response = session.request('GET', url).data.decode()

        description = f""
        soup = BeautifulSoup(response, 'lxml')
        try:
            data = soup.find('div', attrs={'data-test-selector': 'body-content'})
            data_items:list[str] = [li.text.strip() for li in data.find_all('li')] # type: ignore
        except AttributeError:
            return description

        for result in data_items:
            text = f"\u25cf {result}"
            description = description + text + "\n"
        
        return description
    
    try:
        _page_version = f"https://raw.githubusercontent.com/Spheres-cu/subdx-dl/refs/heads/main/sdx_dl/__init__.py"
        _dt_version = session.request('GET', _page_version).data.decode()
        _g_version = _dt_version.split('"')[1]

        if _g_version > version:

            msg = gl("New_version_available") + _g_version + ":\n\n"\
                + get_version_description(_g_version) + "\n"\
                + gl("Please_update_your_current_version") + version +"\r\n"        
        else:
            msg = gl("No_new_version_available") +\
                  gl("Current_version") + version + "\r\n"

    except (HTTPError, Exception) as e:
        ExceptionErrorMessage(e)

    return msg

### Get Remaining arguments
def _get_remain_arg(args: List[str] | str) -> str:
    """ Get remainig arguments values"""
    n = 0; arg = ""
    for i in sys.argv:
        if i in args:
            arg = sys.argv[n + 1] if n + 1 < len(sys.argv) else arg
            break
        n = n + 1
    return arg

### Check version action class
class ChkVersionAction(argparse.Action):
    """Class Check version. This class call for `check_version` function"""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw,):
        super().__init__(nargs=nargs, **kw) 
    
    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):       
        p = getattr(namespace, "proxy") or _get_remain_arg(["-x", "--proxy"])
        if not p:
            config = ConfigManager()
            proxy = config.get("proxy")
        else:
            proxy = p if validate_proxy(p) else None
        
        print(_check_version(version("subdx-dl"), proxy))
        sys.exit(0)

### Class VideoExtractor ###
class VideoMetadataExtractor:
    """
    A class to extract metadata from video filenames using guessit.
    """
    @typing.no_type_check
    @staticmethod
    def extract_all(filename: str, options:str|Dict[str, Any]={}) -> Dict[str, Any]:
        """
        Extract all available metadata from a video filename.
        
        Args:
            filename (str): The video filename to parse
        
        :param options:
        :type options: str|dict
                   
        Returns:
            dict: Dictionary containing all extracted properties
        """
        all_metadata = guessit(filename, options) 
        result_dict:Dict[str, Any] = {}
        for key, value in all_metadata.items(): 
            if isinstance(value, jsonutils.Match):
                result_dict[key] = {
                    'value': value.value, 
                    'raw': value.raw 
                }
            else:
                result_dict[key] = value
        return result_dict
    
    @staticmethod
    def extract_specific(filename: str, *properties: str, options:str|Dict[str, Any]={}) -> Dict[str, Any]:
        """
        Extract specific properties from a video filename.
        
        Args:
            filename (str): The video filename to parse
            *properties (str): Properties to extract (e.g., 'title', 'year')
    
        :param options:
        :type options: str|dict
            
        Returns:
            dict: Dictionary containing only the requested properties
        """
        all_metadata:Dict[str, Any] = {}
        all_metadata = guessit(filename, options)  # type: ignore
        result_dict: Dict[str, Any] = {}
        for key, value in all_metadata.items(): # type: ignore
            if isinstance(value, jsonutils.Match):
                result_dict[key] = {
                    'value': value.value,  # type: ignore
                    'raw': value.raw # type: ignore
                }
            else:
                result_dict[key] = value
        
        return {prop: result_dict.get(prop) for prop in properties}
    
    @staticmethod
    def pretty_print(metadata: Dict[str, Any]) -> None:
        """
        Pretty print the metadata dictionary.
        
        Args:
            metadata (dict): Metadata dictionary to print
        """
        
        console.print_json(data=metadata, indent=4, default=str)

### Class Config Settings
class ConfigManager:
    """
    A class to manage application configuration settings in a JSON file.
    
    Attributes:
        config_path (str): Path to the configuration file
        config (dict): Dictionary containing the configuration settings
    """
    
    def __init__(self, config_path: str = ""):
        """
        Initialize the ConfigManager with a path to the configuration file.
        
        Args:
            config_path (str): Path to the configuration file. Defaults to None.
        """
        self.config_path = config_path if config_path else self.get_path()
        self.config = {}
        
        # Load existing config if it exists
        self._load_config()
        
    @property
    def exists(self) -> bool:
        """ Check if exists a config file"""
        return os.path.isfile(self.config_path)
    
    @property
    def hasconfig(self) -> bool:
        """ Check if config is empty"""
        return bool(self.config)
    
    def _load_config(self) -> None:
        """Load the configuration from file or create a new one if it doesn't exist."""
        try:
            if self.exists:
                with open(self.config_path, 'r') as f:
                    self.config = json.load(f)
            else:
                self.config = {}
        except (json.JSONDecodeError, IOError) as e:
            pass
            console.print(":no_entry:[bold red] " + gl("Failed_to_load_configuration") + "[/]" + f'{e.__class__.__name__}\n',
                    emoji=True, new_line_start=True)
            self._save_config()
            sys.exit(1)

    def _save_config(self) -> None:
        """Save the current configuration to file."""
        if not self.exists:
            config_dir = Path(os.path.dirname(self.config_path))
            config_dir.mkdir(parents=True, exist_ok=True)
        try:
            with open(self.config_path, 'w') as f:
                json.dump(self.config, f, indent=4)
        except IOError as e:
            pass
            console.print(":no_entry:[bold red] " + gl("Failed_to_save_configuration") + "[/]" + f'{e.__class__.__name__}\n',
                    emoji=True, new_line_start=True)
            sys.exit(1)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Get a configuration value by key.
        
        Args:
            key (str): The configuration key to retrieve
            default (Any): Default value to return if key doesn't exist
            
        Returns:
            The configuration value or default if key doesn't exist
        """
        return self.config.get(key, default)

    def set(self, key: str, value: Any) -> None:
        """
        Set a configuration value.
        
        Args:
            key (str): The configuration key to set
            value (Any): The value to set
        """
        self.config[key] = value
        self._save_config()

    def update(self, new_config: Dict[str, Any]) -> None:
        """
        Update multiple configuration values at once.
        
        Args:
            new_config (dict): Dictionary of key-value pairs to update
        """
        self.config.update(new_config)
        self._save_config()

    def delete(self, key: str) -> None:
        """
        Delete a configuration key.
        
        Args:
            key (str): The configuration key to delete
        """
        if key in self.config:
            del self.config[key]
            self._save_config()

    def reset(self) -> None:
        """Reset the configuration to an empty state."""
        self.config = {}
        self._save_config()

    def get_all(self) -> Dict[str, Any]:
        """
        Get all configuration settings.
        
        Returns:
            dict: A copy of the current configuration
        """
        return self.config.copy()

    def save_all(self, config:Dict[str, Any]) -> None:
        """
        Save all configuration values.
        
        Args:
            dict: With all configuration values.
        """

        self.reset()
        self.config = config.copy()
        self._save_config()
       
    def print_config(self) -> None:
        """
        Pretty print the config dictionary.
        """
        console.print_json(data=self.config, indent=4, default=str)

    def merge_config(self, args:Dict[str, Any]) -> Dict[str, Any]:
        """
        Merge args values with config file
        
        Args:
            dict: With arguments to merge
        """
        merged: Dict[str, Any] = {}
        merged = {**args, **{k: v for k, v in self.config.items() if not args[k]}}

        return merged

    @staticmethod
    def get_path(app_name: str = "subdx-dl", file_name: Optional[str] = "sdx-config.json") -> Path:
        """
        Get the appropriate local configuration path for the current platform.
        
        Args:
            app_name: Name of your application (used to create a subdirectory). Default subdx-dl
            file_name: Optional filename to append to the config path. Default sdx-config.json
            
        Returns:
            Path object pointing to the configuration directory or file
            
        Platform-specific paths:
        - Windows: %LOCALAPPDATA%\\<app_name>\\
        - macOS: ~/Library/Application Support/<app_name>/
        - Linux: ~/.config/<app_name>/
        """
        if sys.platform == "win32":
            # Windows
            base_dir = Path(f'{os.getenv("LOCALAPPDATA")}')
        elif sys.platform == "darwin":
            # macOS
            base_dir = Path.home() / "Library" / "Application Support"
        else:
            # Linux and other UNIX-like systems
            base_dir = Path.home() / ".config"
        
        config_dir = base_dir / app_name
        
        if file_name:
            return config_dir / file_name
        return config_dir

### Config action classes
class ViewConfigAction(argparse.Action):
    """Check config file class Action"""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw,): 
        super().__init__(nargs=nargs, **kw) 
    
    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None): 
        config = ConfigManager()
        if config.exists:
            console.print("[bold yellow]" + gl("Config_file") + "[/]", f'{config.get_path()}')
            config.print_config() if config.hasconfig else console.print(":no_entry: [bold red]" + gl("Config_is_empty") + "[/]")
        else:
            console.print(":no_entry: [bold red]" + gl("Not_exists_a_config_file") + "[/]")
        sys.exit(0)

class SaveConfigAction(argparse.Action):
    """Save allowed arguments to a config file. Existing values are update."""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw,): 
        super().__init__(nargs=nargs, **kw) 
   
    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None): 
        allowed_values = ["quiet", "verbose", "force", "no_choose", "no_filter", "nlines", "path", "proxy", "Season", "imdb", "lang"]
        copied_config = namespace.__dict__.copy()

        if all(not copied_config[k] for k in copied_config.keys()):
            console.print(":no_entry:[bold yellow]" + gl("Nothing_to_save") + "[/]")
            sys.exit(0)
  
        for k in namespace.__dict__.keys():
            if k not in allowed_values: del copied_config[k]
        
        config = ConfigManager()

        config.update(config.merge_config(copied_config)) if config.hasconfig else config.save_all(copied_config)
        if not copied_config['quiet']: console.print(":heavy_check_mark: " + gl("Config_was_saved"))
        
        if not getattr(namespace, "search"):
            sys.exit(0)

class SetConfigAction(argparse.Action):
    """Save an option to config file"""
    @typing.no_type_check
    def __init__(self, nargs='?', **kw):
        super().__init__(nargs=nargs, **kw) 
    
    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string = None): 

        if not values:
            console.print(":no_entry:[bold red]  " + gl("Not_a_valid_option") + "[/]", self.choices)
            sys.exit(1)
        
        key, value = f'', None
        config = ConfigManager()

        if values in ["quiet", "verbose", "force", "no_choose", "no_filter", "Season", "imdb"]:
            key, value = f'{values}', bool(True)
        elif values == "path":
            path = _get_remain_arg("path")
            if os.path.isdir(path) and os.access(path, os.W_OK):
                key, value = f'{values}', path
            else:
                console.print(":no_entry:[bold red]" + gl("Directory") + "[/]" + ":[yellow] " + path + "[bold red] "\
                            + gl("Directory_not_exists") + "[/]")
        elif values == "proxy":
            proxy = _get_remain_arg("proxy")
            if validate_proxy(proxy):
                key, value = f'{values}', proxy
            else:
                console.print(":no_entry:[bold red] " + gl("Incorrect_proxy_setting").split('.')[0] + ":[yellow] " + proxy + "[/]")
        elif values == "nlines":
            lines = _get_remain_arg("nlines") 
            key, value = f'{values}', int(lines) if lines.isnumeric() and int(lines) in range(5,25,5) else 10
        elif values == "lang":
            language = _get_remain_arg("lang") 
            key, value = f'{values}', language if language in ["es", "en"] else "es"
        
        if not value:
            sys.exit(1)

        if config.hasconfig:
            config.set(key, value)
        else:
            config.update({key: value})
        
        console.print(gl("Done"))
        sys.exit(0)

class ResetConfigAction(argparse.Action):
    """Reset an option in the config file"""
    @typing.no_type_check
    def __init__(self, nargs='?', **kw): 
        super().__init__(nargs=nargs, **kw) 
    
    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string = None):

        if not values:
            console.print(":no_entry:[bold red]  " + gl("Not_a_valid_option") + "[/]", self.choices)
            sys.exit(1)

        config = ConfigManager()

        if values in ["quiet", "verbose", "force", "no_choose", "no_filter", "Season", "imdb", "path", "proxy", "nlines", "lang"]:
            config.delete(values)
        
        console.print(gl("Done"))
        sys.exit(0)

# Setting config language
config = ConfigManager()
if config.hasconfig and 'lang' in config.config:
    set_locale(config.config['lang'])

### Findfiles class ###
extension_pattern = '(\\.[a-zA-Z0-9]+)$'
string_type = str

class InvalidPath(Exception):
    """Raised when an argument is a non-existent file or directory path
    """
    pass

class FindFiles(object):
    """Given a file, it will verify it exists. Given a folder it will descend
    one level into it and return a list of files, unless the recursive argument
    is True, in which case it finds all files contained within the path.

    The with_extension argument is a list of valid extensions, without leading
    spaces. If an empty list (or None) is supplied, no extension checking is
    performed.

    The filename_blacklist argument is a list of regexp strings to match against
    the filename (minus the extension). If a match is found, the file is skipped
    (e.g. for filtering out "sample" files). If [] or None is supplied, no
    filtering is done
    """

    def __init__(self, path:str, with_extension: Optional[List[str]] = None, filename_blacklist: Optional[List[Any]] = None, recursive:bool = False):
        self.path = path
        if with_extension is None:
            self.with_extension = []
        else:
            self.with_extension = with_extension
        if filename_blacklist is None:
            self.with_blacklist = []
        else:
            self.with_blacklist = filename_blacklist
        self.recursive = recursive
    
    @staticmethod
    def split_extension(filename:str) -> str:
        """Split extension from `filename` based in extension pattern"""
        base = re.sub(extension_pattern, "", filename)
        ext = filename.replace(base, "")
        return ext

    def findFiles(self) -> List[str]:
        """Returns list of files found at path
        """
        listfiles:List[str] = []
        if os.path.isfile(self.path):
            path = os.path.abspath(self.path)
            if self._checkExtension(path) and not self._blacklistedFilename(path):
                listfiles.append(path)
                return listfiles
            else:
                return listfiles
        elif os.path.isdir(self.path):
            return self._findFilesInPath(self.path)
        else:
            raise InvalidPath("%s is not a valid file/directory" % self.path)

    def _checkExtension(self, fname:str) -> bool:
        """Checks if the file extension is blacklisted in valid_extensions
        """
        if len(self.with_extension) == 0:
            return True

        # don't use split_extension here (otherwise valid_extensions is useless)!
        _, extension = os.path.splitext(fname)
        for cext in self.with_extension:
            cext = ".%s" % cext
            if extension == cext:
                return True
        else:
            return False

    def _blacklistedFilename(self, filepath:str) -> bool:
        """Checks if the filename (optionally excluding extension)
        matches filename_blacklist

        self.with_blacklist should be a list of strings and/or dicts:

        a string, specifying an exact filename to ignore
        "filename_blacklist": [".DS_Store", "Thumbs.db"],

        a dictionary, where each dict contains:

        Key 'match' - (if the filename matches the pattern, the filename
        is blacklisted)

        Key 'is_regex' - if True, the pattern is treated as a
        regex. If False, simple substring check is used (if
        cur['match'] in filename). Default is False

        Key 'full_path' - if True, full path is checked. If False, only
        filename is checked. Default is False.

        Key 'exclude_extension' - if True, the extension is removed
        from the file before checking. Default is False.
        """

        if len(self.with_blacklist) == 0:
            return False

        fullname = f'{os.path.split(filepath)[1]}'
        fname = self.split_extension(fullname)

        for fblacklist in self.with_blacklist:
            if isinstance(fblacklist, string_type):
                if fullname == fblacklist:
                    return True
                else:
                    continue

            if "full_path" in fblacklist and fblacklist["full_path"]:
                to_check = filepath
            else:
                if fblacklist.get("exclude_extension", False):
                    to_check = fname
                else:
                    to_check = fullname

            if fblacklist.get("is_regex", False):
                m = re.match(fblacklist["match"], to_check)
                if m is not None:
                    return True
            else:
                m = fblacklist["match"] in to_check
                if m:
                    return True
        else:
            return False

    def _findFilesInPath(self, startpath: str) -> List[str]:
        """Finds files from startpath, could be called recursively
        """
        allfiles:List[str] = []
        if not os.access(startpath, os.R_OK):
            return allfiles

        for subf in os.listdir(string_type(startpath)):
            newpath = os.path.join(startpath, subf)
            newpath = os.path.abspath(newpath)
            if os.path.isfile(newpath):
                if not self._checkExtension(subf):
                    continue
                elif self._blacklistedFilename(subf):
                    continue
                else:
                    allfiles.append(newpath)
            else:
                if self.recursive:
                    allfiles.extend(self._findFilesInPath(newpath))
        return allfiles


# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import time
import shutil
import tempfile
from tempfile import NamedTemporaryFile
from rarfile import RarCannotExec, RarExecError # type: ignore
from rarfile import RarFile, is_rarfile # type: ignore
from zipfile import ZipFile, is_zipfile
from sdx_dl.sdxparser import args, logger
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl
from typing import Dict, Any
from sdx_dl.sdxutils import extract_subtitles # type: ignore
from sdx_dl.sdxutils import (get_imdb_search, get_aadata, convert_date, get_filtered_results, sort_results, get_selected_subtitle_id,
HTTPErrorsMessageException, clean_screen, paginate, Metadata, metadata, SUBDIVX_DOWNLOAD_PAGE, HTTPError, headers, conn)

__all__ = ["get_subtitle_id", "get_subtitle"]

def get_subtitle_id(title:str, number:str, inf_sub:Dict[str, Any], metadata: Metadata=metadata):
    
    """
    Get a list of subtitles of subtitles searched by ``title`` and season/episode
    ``number`` of series or movies.
      
    The results are ordered based on a weighing of a ``metadata`` list.

    If ``no_choose`` ``(-nc)`` is false then a list of subtitles is show for choose.

    Else the first founded subtitle `id` is choosen.

    Return the subtitle `id`
    """
    buscar = None
    res = ""
    list_Subs_Dicts:list[Dict[str,Any]] = []
    
    if args.imdb:
        if not args.quiet:
            console.print(":earth_americas: [bold yellow]" + gl("Searching_IMDB") +  f"{title} {number}", new_line_start=True, emoji=True) 
        logger.debug(f'Searching in IMDB: {str(title + " " + number)}')
        search = get_imdb_search(title, number, inf_sub)
        buscar = search
        if buscar is not None and inf_sub['type'] == 'episode':
            title = buscar.replace(number, "").strip()
        logger.debug(f'IMDB Search result:{buscar}')

        if not args.quiet:
            clean_screen()
            imdb_search = buscar if buscar is not None else "Ninguno"
            console.print(":information_source: [bold yellow] " + gl("Search_terms_from_IMDB") + "[/]" + imdb_search, new_line_start=True, emoji=True)
            time.sleep(0.5)

    if buscar is None : buscar = f"{title} {number}".strip()

    if not args.quiet:console.print("\r")
    logger.debug(f'Searching subtitles for: ' + str(title) + " " + str(number).upper())
    
    with console.status(gl("Searching_subtitles_for") + str(title) + " " + str(number).upper()) as status:
        status.start() if not args.quiet else status.stop()
        json_aaData = get_aadata(buscar)
 
    if json_aaData["iTotalRecords"] == 0 :
        if not args.quiet: console.print(":no_entry: [bold red]" + gl("Not_subtitles_records_found_for") + "[/]" + "[yellow]" + buscar +"[/]")
        logger.debug(f'Not subtitles records found for: "{buscar}"')
        return res
    else:
        logger.debug(f'Found subtitles records for: "{buscar}"')
    
    # Get Json Data Items
    aaData_Items = json_aaData['aaData']
    
    if aaData_Items is not None:
        list_Subs_Dicts = convert_date(aaData_Items)
    else:
        if not args.quiet: console.print(":no_entry: [bold red] " + gl("No_suitable_data_were_found_for") + "[yellow]" + buscar +"[/]")
        logger.debug(f'No suitable data were found for: "{buscar}"')
        return res
    
    # only include results for this specific serie / episode
    # ie. search terms are in the title of the result item
    
    if args.imdb or args.no_filter:
        filtered_list_Subs_Dicts = list_Subs_Dicts
    else:
        filtered_list_Subs_Dicts = get_filtered_results(title, number, inf_sub, list_Subs_Dicts)

    if not filtered_list_Subs_Dicts:
        if not args.quiet: console.print(":no_entry: [bold red]" + gl("No_suitable_data_were_found_for") +  "[yellow]" + buscar +"[/]")
        logger.debug(f'No suitable data were found for: "{buscar}"')
        return res
    
    if metadata.hasdata:
        results = sort_results(filtered_list_Subs_Dicts, metadata)
    else:
        results = sorted(filtered_list_Subs_Dicts, key=lambda item: (item['descargas']), reverse=True)

    # Print subtitles search infos
    # Construct Table for console output
    
    table_title = str(title) + " " + str(number).upper()
    results_pages = paginate(results, 10)

    if (args.no_choose == False):
        res = get_selected_subtitle_id(table_title, results, metadata)
        return res
    else:
        # get first subtitle
        res = f"{results_pages['pages'][0][0]['id']}"
    
    return res

def get_subtitle(subid:str, topath:str):
    """Download a subtitle with id ``subid`` to a destination ``path``."""

    url = f"{SUBDIVX_DOWNLOAD_PAGE + 'descargar.php?id=' + subid}"
    
    if not args.quiet: clean_screen()
    temp_file = NamedTemporaryFile(delete=False)

    # get direct download link
    if not args.quiet: console.print(gl("Downloading_Subtitle"), emoji=True,new_line_start=True)
    logger.debug(f"Trying Download from link: {url}")
    try:
        download_url = conn.request('GET', url, headers=headers)
    except HTTPError as e:
        HTTPErrorsMessageException(e)
        sys.exit(1)

    if download_url:
        logger.debug(f"Downloaded from: {SUBDIVX_DOWNLOAD_PAGE}{download_url.geturl()}")
        temp_file.write(download_url.data)
        temp_file.seek(0)
        # Checking if the file is zip or rar then decompress
        try:
            if is_zipfile(temp_file):
                compressed_sub_file = ZipFile(temp_file)
                extract_subtitles(compressed_sub_file, topath)
            elif is_rarfile(temp_file):
                compressed_sub_file = RarFile(temp_file)
                extract_subtitles(compressed_sub_file, topath)
        except (RarCannotExec, RarExecError):
                console.clear()
                temp_dir = tempfile.gettempdir()
                shutil.copyfile(os.path.join(temp_dir, temp_file.name), os.path.join(topath, f'{subid}.rar')) 

                console.print(":warning: [bold red] " + gl("Cannot_find_a_working_tool") + "[bold yellow] " + gl("Install_rar")\
                            + "[/]", emoji=True, new_line_start=True)
                logger.debug(f"Cannot find a working tool, please install rar decompressor tool")
                logger.debug(f"File downloaded to: {os.path.join(topath, f'{subid}.rar')}")
    else:
        temp_file.close()
        os.unlink(temp_file.name)
        logger.error(f'No suitable subtitle download for : "{url}"')
        if not args.quiet: console.print(":cross_mark: [bold red]" + gl("No_suitable_subtitle_to_download") + "[/]",emoji=True, new_line_start=True)
        sys.exit(1)
        time.sleep(2)
            
    # Cleaning
    temp_file.close()
    os.unlink(temp_file.name)



# Copyright (C) 2025 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from importlib.resources import files
from typing import Dict, Optional, Any, Set
from locale import getlocale
from sdx_dl.sdxconsole import console
import json
import sys

local_language = f"{getlocale()[0]}".split('_')[0] if getlocale()[0] else ""

__all__ = ['Translator', 'set_locale', 'get_locale', 'gl']

class Translator:
    def __init__(self, locale: str = local_language):
        """
        Initialize the translator with a default locale.
        
        Args:
            locale: Default language code ('es', 'en')
        """
        self.locale = locale if files('sdx_dl.language').joinpath(f"{locale}.json").is_file() else "en"
        self._translations: Dict[str, Dict[str, str]] = {}
        self._loaded_locales:Set[str] = set()

    def load_translations(self, locale: str) -> None:
        """
        Load translations for a specific locale if not already loaded.
        
        Args:
            locale: Language code to load
        """
        if locale in self._loaded_locales:
            return
        
        if files('sdx_dl.language').joinpath(f"{self.locale}.json").is_file():
            data_file = files('sdx_dl.language').joinpath(f"{self.locale}.json").open(mode='r', encoding='utf-8')
            try: 
                self._translations[locale] = json.load(data_file)
                self._loaded_locales.add(locale)
            except json.JSONDecodeError:
                raise ValueError(f"Invalid JSON in translation file for locale: {locale}")
        else:
            console.print(":no_entry: [bold red] Don't exists a language file!")
            sys.exit(1)

    def set_locale(self, locale: str) -> None:
        """
        Set the current locale and load its translations if not already loaded.
        
        Args:
            locale: Language code to set as current
        """
        self.locale = locale
        self.load_translations(locale)

    def translate(self, key: str, locale: Optional[str] = None, **kwargs: Any) -> str:
        """
        Translate a string key to the specified or current locale.
        
        Args:
            key: Translation key
            locale: Optional language code (uses current locale if None)
            **kwargs: Formatting arguments for the translated string
            
        Returns:
            Translated string, or the key if translation not found
        """
        locale = locale or self.locale
        if locale not in self._loaded_locales:
            self.load_translations(locale)
            
        translations = self._translations.get(locale, {})
        translated = translations.get(key, key)
        
        if kwargs:
            try:
                return translated.format(**kwargs)
            except (KeyError, ValueError):
                return translated
        return translated


# Singleton instance for easy access
translator = Translator()

def set_locale(locale: str) -> None:
    """Set the locale for the default translator instance."""
    translator.set_locale(locale)

def get_locale(key: str, **kwargs: Any) -> str:
    """Translate a key using the default translator instance."""
    return translator.translate(key, **kwargs)

# Alias for get_locale
gl = get_locale